import warnings
import numpy as np
import inspect

from statsmodels.tsa import stattools
from statsmodels.tsa.vector_ar.vecm import coint_johansen
from sklearn.gaussian_process import kernels, GaussianProcessRegressor
from sklearn.metrics import mean_squared_error
from sklearn import linear_model
import mne.connectivity as mnec

from pyspi.base import (
    Directed,
    Undirected,
    Unsigned,
    parse_bivariate,
    parse_multivariate,
)


class Cointegration(Undirected, Unsigned):

    name = "Cointegration"
    identifier = "coint"
    labels = ["misc", "unsigned", "temporal", "undirected", "nonlinear"]

    def __init__(
        self,
        method="johansen",
        statistic="trace_stat",
        det_order=1,
        k_ar_diff=1,
        autolag="aic",
        maxlag=10,
        trend="c",
    ):
        self._method = method
        self._statistic = statistic
        if method == "johansen":
            self.identifier += (
                f"_{method}_{statistic}_order-{det_order}_ardiff-{k_ar_diff}"
            )
            self._det_order = det_order
            self._k_ar_diff = k_ar_diff
        else:
            self._autolag = autolag
            self._maxlag = maxlag
            self._trend = trend
            self.identifier += (
                f"_{method}_{statistic}_trend-{trend}_autolag-{autolag}_maxlag-{maxlag}"
            )

    @property
    def key(self):
        key = (self._method,)
        if self._method == "johansen":
            return key + (self._det_order, self._k_ar_diff)
        else:
            return key + (self._autolag, self._maxlag, self._trend)

    def _from_cache(self, data, i, j):
        idx = (i, j)
        try:
            ci = data.coint[self.key][idx]
        except (KeyError, AttributeError):
            z = data.to_numpy(squeeze=True)

            if self._method == "aeg":
                stats = stattools.coint(
                    z[i],
                    z[j],
                    autolag=self._autolag,
                    maxlag=self._maxlag,
                    trend=self._trend,
                )

                ci = {"tstat": stats[0]}
            else:
                stats = coint_johansen(
                    z[[i, j]].T, det_order=self._det_order, k_ar_diff=self._k_ar_diff
                )

                ci = {
                    "max_eig_stat": stats.max_eig_stat[0],
                    "trace_stat": stats.trace_stat[0],
                }

            try:
                data.coint[self.key][idx] = ci
            except AttributeError:
                data.coint = {self.key: {idx: ci}}
            except KeyError:
                data.coint[self.key] = {idx: ci}
            data.coint[self.key][(j, i)] = ci

        return ci

    # Return the negative t-statistic (proxy for how co-integrated they are)
    @parse_bivariate
    def bivariate(self, data, i=None, j=None, verbose=False):
        ci = self._from_cache(data, i, j)
        return ci[self._statistic]


class LinearModel(Directed, Unsigned):
    name = "Linear model regression"
    identifier = "lmfit"
    labels = ["misc", "unsigned", "unordered", "normal", "linear", "directed"]

    def __init__(self, model):
        self.identifier += f"_{model}"
        self._model = getattr(linear_model, model)

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model_params = inspect.signature(self._model).parameters
            if "random_state" in model_params:
                mdl = self._model(random_state=42).fit(z[i], np.ravel(z[j]))
            else:
                mdl = self._model().fit(z[i], np.ravel(z[j]))
        y_predict = mdl.predict(z[i])
        return mean_squared_error(y_predict, np.ravel(z[j]))


class GPModel(Directed, Unsigned):
    name = "Gaussian process regression"
    identifier = "gpfit"
    labels = ["misc", "unsigned", "unordered", "normal", "nonlinear", "directed"]

    def __init__(self, kernel="RBF"):
        self.identifier += f"_{kernel}"
        self._kernel = kernels.ConstantKernel() + kernels.WhiteKernel()
        self._kernel += getattr(kernels, kernel)()

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            gp = GaussianProcessRegressor(kernel=self._kernel).fit(z[i], np.ravel(z[j]))
        y_predict = gp.predict(z[i])
        return mean_squared_error(y_predict, np.ravel(z[j]))


class PowerEnvelopeCorrelation(Undirected, Unsigned):
    humanname = "Power envelope correlation"
    identifier = "pec"
    labels = ["unsigned", "misc", "undirected"]

    def __init__(self, orth=False, log=False, absolute=False):
        self._orth = False
        if orth:
            self._orth = "pairwise"
            self.identifier += "_orth"
        self._log = log
        if log:
            self.identifier += "_log"
        self._absolute = absolute
        if absolute:
            self.identifier += "_abs"

    @parse_multivariate
    def multivariate(self, data):
        z = np.moveaxis(data.to_numpy(), 2, 0)
        adj = np.squeeze(
            mnec.envelope_correlation(
                z, orthogonalize=self._orth, log=self._log, absolute=self._absolute
            )
        )
        np.fill_diagonal(adj, np.nan)
        return adj



import numpy as np
import pandas as pd
from cdt.causality.pairwise import ANM, CDS, IGCI, RECI
import pyEDM
from sklearn.gaussian_process import GaussianProcessRegressor
from cdt.causality.pairwise.ANM import normalized_hsic

from pyspi.base import Directed, Unsigned, Signed, parse_bivariate, parse_multivariate


class AdditiveNoiseModel(Directed, Unsigned):

    name = "Additive noise model"
    identifier = "anm"
    labels = ["unsigned", "causal", "unordered", "linear", "directed"]
    
    # monkey-patch the anm_score function, see cdt PR #155
    def corrected_anm_score(self, x, y):
        gp = GaussianProcessRegressor(random_state=42).fit(x, y)
        y_predict = gp.predict(x).reshape(-1, 1) 
        indepscore = normalized_hsic(y_predict - y, x)
        return indepscore
    
    ANM.anm_score = corrected_anm_score

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy()
        return ANM().anm_score(z[i], z[j])


class ConditionalDistributionSimilarity(Directed, Unsigned):

    name = "Conditional distribution similarity statistic"
    identifier = "cds"
    labels = ["unsigned", "causal", "unordered", "nonlinear", "directed"]

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy()
        return CDS().cds_score(z[i], z[j])


class RegressionErrorCausalInference(Directed, Unsigned):

    name = "Regression error-based causal inference"
    identifier = "reci"
    labels = ["unsigned", "causal", "unordered", "nonlinear", "directed"]

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy()
        return RECI().b_fit_score(z[i], z[j])


class InformationGeometricConditionalIndependence(Directed, Unsigned):

    name = "Information-geometric conditional independence"
    identifier = "igci"
    labels = ["causal", "directed", "nonlinear", "unsigned", "unordered"]

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy()
        return IGCI().predict_proba((z[i], z[j]))


class ConvergentCrossMapping(Directed, Signed):

    name = "Convergent cross-mapping"
    identifier = "ccm"
    labels = ["causal", "directed", "nonlinear", "temporal", "signed"]

    def __init__(self, statistic="mean", embedding_dimension=None):
        self._statistic = statistic
        self._E = embedding_dimension

        self.identifier += f"_E-{embedding_dimension}_{statistic}"

    @property
    def key(self):
        return self._E

    def _from_cache(self, data):
        try:
            ccmf = data.ccm[self.key]
        except (AttributeError, KeyError):
            z = data.to_numpy(squeeze=True)

            M = data.n_processes
            N = data.n_observations
            df = pd.DataFrame(
                np.concatenate([np.atleast_2d(np.arange(0, N)), z]).T,
                columns=["index"] + [f"proc{p}" for p in range(M)],
            )

            # Get the embedding
            if self._E is None:
                embedding = np.zeros((M, 1))

                # Infer optimal embedding from simplex projection
                for _i in range(M):
                    pred = str(10) + " " + str(N - 10)
                    embed_df = pyEDM.EmbedDimension(
                        dataFrame=df,
                        lib=pred,
                        pred=pred,
                        columns=df.columns.values[_i + 1],
                        showPlot=False,
                    )
                    embedding[_i] = embed_df.max()["E"]
            else:
                embedding = np.array([self._E] * M)

            # Compute CCM from the fixed or optimal embedding
            nlibs = 21
            ccmf = np.zeros((M, M, nlibs + 1))
            for _i in range(M):
                for _j in range(_i + 1, M):
                    try:
                        E = int(max(embedding[[_i, _j]]))
                    except NameError:
                        E = int(self._E)

                    # Get list of library sizes given nlibs and lower/upper bounds based on embedding dimension
                    upperE = int(np.floor((N - E - 1) / 10) * 10)
                    lowerE = int(np.ceil(2 * E / 10) * 10)
                    inc = int((upperE - lowerE) / nlibs)
                    lib_sizes = str(lowerE) + " " + str(upperE) + " " + str(inc)
                    srcname = df.columns.values[_i + 1]
                    targname = df.columns.values[_j + 1]
                    ccm_df = pyEDM.CCM(
                        dataFrame=df,
                        E=E,
                        columns=srcname,
                        target=targname,
                        libSizes=lib_sizes,
                        sample=100,
                        seed=42,
                    )
                    ccmf[_i, _j] = ccm_df.iloc[:, 1].values[: (nlibs + 1)]
                    ccmf[_j, _i] = ccm_df.iloc[:, 2].values[: (nlibs + 1)]

            try:
                data.ccm[self.key] = ccmf
            except AttributeError:
                data.ccm = {self.key: ccmf}
        return ccmf

    @parse_multivariate
    def multivariate(self, data):
        ccmf = self._from_cache(data)

        if self._statistic == "mean":
            return np.nanmean(ccmf, axis=2)
        elif self._statistic == "max":
            return np.nanmax(ccmf, axis=2)
        elif self._statistic == "diff":
            return np.nanmean(ccmf - np.transpose(ccmf, axes=[1, 0, 2]), axis=2)
        else:
            raise TypeError(f"Unknown statistic: {self._statistic}")



from pyspi.utils import filter_spis
import pytest
import yaml
from unittest.mock import mock_open, patch

@pytest.fixture
def mock_yaml_content():
    return {
        "module1": {
            "spi1": {"labels": ["keyword1", "keyword2"], "configs": [1, 2]},
            "spi2": {"labels": ["keyword1"], "configs": [3]},
        },
        "module2": {
            "spi3": {"labels": ["keyword3"], "configs": [1, 2, 3]},
        },
    }

def test_filter_spis_invalid_keywords():
    """Pass in a dataype other than a list for the keywords"""
    with pytest.raises(ValueError) as excinfo:
        filter_spis(keywords="linear", configfile="pyspi/config.yaml")
    assert "Keywords must be provided as a list of strings" in str(excinfo.value)
    # check for passing in an empty list
    with pytest.raises(ValueError) as excinfo:
        filter_spis(keywords=[], configfile="pyspi/config.yaml")
    assert "At least one keyword must be provided" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        filter_spis(keywords=[4], configfile="pyspi/config.yaml")
    assert "All keywords must be strings" in str(excinfo.value)  

def test_filter_spis_with_invalid_config():
    """Pass in an invalid/missing config file"""
    with pytest.raises(FileNotFoundError):
        filter_spis(keywords=["test"], configfile="invalid_config.yaml")

def test_filter_spis_no_matches(mock_yaml_content):
    """Pass in keywords that return no spis and check for ValuError"""
    m = mock_open()
    m().read.return_value = yaml.dump(mock_yaml_content)
    keywords = ["random_keyword"]

    with patch("builtins.open", m), \
        patch("os.path.isfile", return_value=True), \
        patch("yaml.load", return_value=mock_yaml_content):
        with pytest.raises(ValueError) as excinfo:
            filter_spis(keywords=keywords, output_name="mock_filtered_config", configfile="./mock_config.yaml")

    assert "0 SPIs were found" in str(excinfo.value), "Incorrect error message returned when no keywords match found."

def test_filter_spis_normal_operation(mock_yaml_content):
    """Test whether the filter spis function works as expected"""
    m = mock_open()
    m().read_return_value = yaml.dump(mock_yaml_content)
    keywords = ["keyword1", "keyword2"] # filter keys
    expected_output_yaml = {
        "module1": {
            "spi1": {"labels": ["keyword1", "keyword2"], "configs": [1,2]}
        }
    }

    with patch("builtins.open", m), patch("os.path.isfile", return_value=True), \
         patch("yaml.load", return_value=mock_yaml_content), \
         patch("yaml.dump") as mock_dump:
        
        filter_spis(keywords=keywords, output_name="mock_filtered_config", configfile="./mock_config.yaml")

        mock_dump.assert_called_once()
        args, _ = mock_dump.call_args # get call args for dump and intercept
        actual_output = args[0]  # the first argument to yaml.dump should be the yaml

        assert actual_output == expected_output_yaml, "Expected filtered YAML does not match actual filtered YAML."

def test_filter_spis_io_error_on_read():
    # check to see whether io error is raised when trying to access the configfile
    with patch("builtins.open", mock_open(read_data="data")) as mocked_file:
        mocked_file.side_effect = IOError("error")
        with pytest.raises(IOError):
            filter_spis(["keyword"], "output", "config.yaml")
            
def test_filter_spis_saves_with_random_name_if_no_name_provided(mock_yaml_content):
    # mock os.urandom to return a predictable name
    random_bytes = bytes([1, 2, 3, 4])
    expected_random_part = "01020304"

    with patch("builtins.open", mock_open()) as mocked_file, patch("os.path.isfile", return_value=True), \
         patch("yaml.load", return_value=mock_yaml_content), patch("os.urandom", return_value=random_bytes):
        
        # run the filter function without providing an output name
        filter_spis(["keyword1"])

        # construct the expected output name
        expected_file_name_pattern = f"config_{expected_random_part}.yaml"

        # check the mocked open function to see if file with expected name is opened (for writing)
        call_args_list = mocked_file.call_args_list
        found_expected_call = any(
            expected_file_name_pattern in call_args.args[0] and
            ('w' in call_args.args[1] if len(call_args.args) > 1 else 'w' in call_args.kwargs.get('mode', ''))
            for call_args in call_args_list
        )

        assert found_expected_call, f"no file with the expected name {expected_file_name_pattern} was saved."

def test_loads_default_config_if_no_config_specified(mock_yaml_content):
    script_dir = "/fake/script/directory"
    default_config_path = f"{script_dir}/config.yaml"

    with patch("builtins.open", mock_open()) as mocked_open, \
         patch("os.path.isfile", return_value=True), \
         patch("yaml.load", return_value=mock_yaml_content), \
         patch("os.path.dirname", return_value=script_dir), \
         patch("os.path.abspath", return_value=script_dir):
        
        # run filter func without specifying a config file 
        filter_spis(["keyword1"])

        # ensure the mock_open was called with the expected path
        assert any(call.args[0] == default_config_path for call in mocked_open.mock_calls), \
        "Expected default config file to be opened."



from pyspi.calculator import Calculator, Data, CalculatorFrame
from pyspi.data import load_dataset
import numpy as np 
import os
import pytest

############################# Test Calculator Object ########################
def test_whether_calculator_instantiates():
    """Basic test to check whether or not the calculator will instantiate."""
    calc = Calculator()
    assert isinstance(calc, Calculator), "Calculator failed to instantiate."

def test_whether_calculator_computes():
    # check whether the calculator runs
    data = np.random.randn(3, 100)
    calc = Calculator(dataset=data)
    calc.compute()

def test_whether_calc_instantiates_without_octave():
    # set octave to false to emulate a system without octave (i.e., fails the check)
    Calculator._optional_dependencies['octave'] = False
    calc = Calculator()
    is_initialised = isinstance(calc, Calculator)
    Calculator._optional_dependencies = {}
    assert is_initialised, "Calculator failed to instantiate without Octave."

def test_whether_calc_instantiates_without_java():
    # set java to false and all other deps to true
    Calculator._optional_dependencies['java'] = False
    Calculator._optional_dependencies['octave'] = True
    calc = Calculator()
    is_initialised = isinstance(calc, Calculator)
    Calculator._optional_dependencies = {}
    assert is_initialised, "Calculator failed to instantiate without Java."

def test_whether_calc_instantiates_wo_optional_deps():
    # set all optional deps to false
    Calculator._optional_dependencies['java'] = False
    Calculator._optional_dependencies['octave'] = False
    calc = Calculator()
    is_initialised = isinstance(calc, Calculator)
    Calculator._optional_dependencies = {}
    assert is_initialised, "Calculator failed to instantiate without optional dependencies."

@pytest.mark.parametrize("subset", [
    'fabfour',
    'fast',
    'sonnet'
])
def test_whether_calculator_instantiates_with_subsets(subset):
    """Test whether the calculator instantiates with each of the available subsets"""
    calc = Calculator(subset=subset)
    assert isinstance(calc, Calculator), "Calculator failed to instantiate"

def test_whether_invalid_subset_throws_error():
    """Test whether the calculator fails to instantiate with an invalid subset"""
    with pytest.raises(ValueError) as excinfo:
        calc = Calculator(subset='nviutw')
    assert "Subset 'nviutw' does not exist" in str(excinfo.value), "Subset not found error not displaying."

def test_whether_calculator_compute_fails_with_no_dataset():
    """Test whether the calculator fails to compute SPIs when no dataset is provided."""
    calc = Calculator()
    with pytest.raises(AttributeError) as excinfo:
        calc.compute()
    assert "Dataset not loaded yet" in str(excinfo.value), "Dataset not loaded yet error not displaying."

def test_calculator_name():
    """Test whether the calculator name is retrieved correctly."""
    calc = Calculator(name="test name")
    assert calc.name == "test name", "Calculator name property did not return the expected string 'test name'"

def test_calculator_labels():
    """Test whether the calculator labels are retreived correctly, when provided."""
    test_labels = ['label1', 'label2']
    calc = Calculator(labels = test_labels)
    assert calc.labels == ['label1', 'label2'], f"Calculator labels property did not return the expected list: {test_labels} "

def test_pass_single_integer_as_dataset():
    """Test whether correct error is thrown when incorrect data type passed into calculator."""
    with pytest.raises(TypeError) as excinfo:
        calc = Calculator(dataset=42)
    assert "Unknown data type" in str(excinfo.value), "Incorrect data type error not displaying for integer dataset."

def test_pass_incorrect_shape_dataset_into_calculator():
    """Test whether an error is thrown when incorrect dataset shape is passed into calculator."""
    dataset_with_wrong_dim = np.random.randn(3, 5, 10)
    with pytest.raises(RuntimeError) as excinfo:
        calc = Calculator(dataset=dataset_with_wrong_dim)
    assert "Data array dimension (3)" in str(excinfo.value), "Incorrect dimension error message not displaying for incorrect shape dataset."

@pytest.mark.parametrize("nan_loc, expected_output", [
    ([1], "[1]"),
    ([1, 2], "[1 2]"),
    ([0, 2, 3], "[0 2 3]")
    ])
def test_pass_dataset_with_nan_into_calculator(nan_loc, expected_output):
    """Check whether ValueError is raised when a dataset containing a NaN is passed into the calculator object"""
    base_dataset = np.random.randn(5, 100)
    for loc in nan_loc:
        base_dataset[loc, 0] = np.nan
    with pytest.raises(ValueError) as excinfo:
        calc = Calculator(dataset=base_dataset)
    assert f"non-numerics (NaNs) in processes: {expected_output}" in str(excinfo), "NaNs not detected in dataset when loading into Calculator!"

def test_pass_dataset_with_inf_into_calculator():
    """Check whether ValueError is raised when a dataset containing an inf/-inf value is passed into the calculator object"""
    base_dataset = np.random.randn(5, 100)
    base_dataset[0, 1] = np.inf
    base_dataset[2, 2] = -np.inf
    with pytest.raises(ValueError) as excinfo:
        calc = Calculator(dataset=base_dataset)
    assert f"non-numerics (NaNs) in processes: [0 2]" in str(excinfo), "NaNs not detected in dataset when loading into Calculator!"

@pytest.mark.parametrize("shape, n_procs_expected, n_obs_expected", [
    ((2, 23), 2, 23),
    ((5, 4), 5, 4),
    ((100, 32), 100, 32)
])
def test_data_object_process_and_observations(shape, n_procs_expected, n_obs_expected):
    """Test whether the number of processes and observations for a given dataset is correct"""
    dat = np.random.randn(shape[0], shape[1])
    calc = Calculator(dataset=dat)
    assert calc.dataset.n_observations == n_obs_expected, f"Number of observations returned by Calculator ({calc.dataset.n_observations}) does not match exepected: {n_obs_expected}"
    assert calc.dataset.n_processes == n_procs_expected, f"Number of processes returned by Calculator ({calc.dataset.n_processes}) does not match exepected: {n_procs_expected}"

@pytest.mark.parametrize("yaml_filename", [
    'fabfour_config', 
    'fast_config',
    'sonnet_config'])
def test_whether_config_files_exist(yaml_filename):
    """Check whether the config, fabfour, fast, sonnet_config files exist"""
    expected_file = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyspi', f'{yaml_filename}.yaml'))
    assert os.path.isfile(expected_file), f"{yaml_filename}.yaml file was not found."

@pytest.mark.parametrize("subset, procs, obs", [
    ("all", 2, 100),
    ("all", 5, 100),
    ("fabfour", 8, 100),
    ("fast", 10, 100),
    ("sonnet", 3, 100)
])
def test_whether_table_shape_correct_before_compute(subset, procs, obs):
    """Test whether the pre-configured table is the correct shape prior to computing SPIs."""
    dat = np.random.randn(procs, obs)
    calc = Calculator(dataset=dat, subset=subset)
    num_spis = calc.n_spis
    expected_table_shape = (procs, num_spis*procs)
    assert calc.table.shape == expected_table_shape, f"Calculator table ({subset}) shape: ({calc.table.shape}) does not match expected shape: {expected_table_shape}"

############################# Test Data Object ########################
def test_data_object_has_been_converted_to_numpyfloat64():
    """Test whether the data object converts passed dataset to numpy array by default."""
    dat = np.random.randn(5, 10)
    calc = Calculator(dataset=dat)
    assert calc.dataset.data_type == np.float64, "Dataset was not converted into a numpy array when loaded into Calculator."

def test_whether_data_instantiates():
    """Test whether the data object instantiates without issue."""
    data_obj = Data()
    assert isinstance(data_obj, Data), "Data object failed to instantiate!"

def test_whether_data_throws_error_when_retrieving_nonexistent_dataset():
    """Test whether the data object throws correct message when trying to access a non-existent dataset."""
    data_obj = Data()
    with pytest.raises(AttributeError) as excinfo:
        dataset = data_obj.data
    assert "'Data' object has no attribute 'data'" in str(excinfo), "Unexpected error message when trying to retrieve non-existent dataset!"

def test_whether_data_throws_error_when_incorrect_dataset_type():
    """Test if correct message is shown when passing invalid dataset data type into data object."""
    with pytest.raises(TypeError) as excinfo:
        d = Data(data=3)
    assert f"Unknown data type" in str(excinfo), "Incorrect error message thrown when invalid dataset loaded into data object."

@pytest.mark.parametrize("order, shape, n_procs_expected, n_obs_expected", [
    ("ps", (3, 100), 3, 100),
    ("sp", (100, 3), 3, 100)
])
def test_whether_dim_order_works(order, shape, n_procs_expected, n_obs_expected):
    """Check that ps and sp correctly specify order of process/obseravtions"""
    dataset = np.random.randn(shape[0], shape[1])
    d = Data(data=dataset, dim_order=order)
    assert d.n_processes == n_procs_expected, f"Number of processes does not match expected for specified dim order: {order}"
    assert d.n_observations == n_obs_expected, f"Number of observations does not match expected for specified dim order: {order}"

def test_whether_data_name_assigned_only_with_dataset():
    """If no dataset is provided, there is no name for the data object (N/A)"""
    d = Data(name='test')
    assert d.name == 'N/A', "Data object name is not N/A when no dataset provided."

def test_whether_data_object_has_name_with_dataset():
    """If dataset is provided, the name will be returned"""
    dataset = np.random.randn(4, 100)
    d = Data(data=dataset, name='test')
    assert d.name == "test", f"Data object name 'test' is not being returned. Instead, {d.name} is returned."

def test_whether_data_normalise_works():
    """Check whether the data is being normalised by default when loading into data object"""
    dataset = 4 * np.random.randn(10, 500)
    d = Data(data=dataset, normalise=True)
    returned_dataset = d.to_numpy(squeeze=True)
    assert returned_dataset.mean() == pytest.approx(0, 1e-8), f"Returned dataset mean is not close to zero: {returned_dataset.mean()}"
    assert returned_dataset.std() == pytest.approx(1, 0.01), f"Returned dataset std is not close to one: {returned_dataset.std()}"

def test_whether_set_data_works():
    """Check whether existing dataset is overwritten by new dataset"""
    old_dataset = np.random.randn(1, 100)
    d = Data(data=old_dataset) # start with empty data object
    new_dataset = np.random.randn(5, 100)
    d.set_data(data=new_dataset)
    # just check the shapes since new datast will be normalised and not equal to the dataset passed in
    assert d.to_numpy(squeeze=True).shape[0] == 5, "Unexpected dataset returned when overwriting existing dataset!"

def test_add_univariate_process_to_existing_data_object():
    # start with initial data object
    dataset = np.random.randn(5, 100)
    orig_data_object = Data(data=dataset)
    # now add additional proc to existing data object
    new_univariate_proc = np.random.randn(1, 100)
    orig_data_object.add_process(proc=new_univariate_proc)
    assert orig_data_object.n_processes == 6, "New dataset number of processes not equal to expected number of processes."

def test_add_multivariate_process_to_existing_data_object():
    """Should not work, can only add univariate process with add_process function"""
    dataset = np.random.randn(5, 100)
    orig_data_object = Data(data=dataset)
    # now add additional procs to existing data object
    new_multivariate_proc = np.random.randn(2, 100)
    with pytest.raises(TypeError) as excinfo:
        orig_data_object.add_process(proc=new_multivariate_proc)
    assert "Process must be a 1D numpy array" in str(excinfo.value), "Expected 1D array error NOT thrown."

# @pytest.mark.parametrize("index", 
#                          [[1], [1, 3], [1, 2, 3]])
# def test_remove_valid_process_from_existing_dataset(index):
#     """Try to remove valid processes from existing dataset by specifying one or more indices. 
#     Check if correct indices are being used."""
#     dataset = np.random.randn(5, 100)
#     d = Data(data=dataset, normalise=False)
#     rows_to_remove = index
#     expected_dataset = np.delete(dataset, rows_to_remove, axis=0)
#     d.remove_process(index)
#     out = d.to_numpy(squeeze=True)
#     assert out.shape[0] == (5 - len(index)), f"Dataset shape after removing {len(index)} proc(s) not equal to {(5 - len(index))}"
#     assert np.array_equal(expected_dataset, out), f"Expected dataset after removing proc(s): {index} not equal to dataset returned."

@pytest.mark.parametrize("dataset_name", ["forex", "cml"])
def test_load_valid_dataset(dataset_name):
    """Test whether the load_dataset function will load all available datasets."""
    dataset = load_dataset(dataset_name)
    assert isinstance(dataset, Data), f"Could not load dataset: {dataset_name}"

def test_load_invalid_dataset():
    """Test whether the load_dataset function throws the correct error/message when trying to load non-existent dataset."""
    with pytest.raises(NameError) as excinfo:
        dataset = load_dataset(name="test")
    assert "Unknown dataset: test" in str(excinfo.value), "Did not get expected error when loading invalid dataset."

def test_calculator_frame_normal_operation():
    """Test whether the calculator frame instantiates as expected."""
    datasets = [np.random.randn(3, 100) for _ in range(3)]
    dataset_names = ['d1', 'd2', 'd3']
    dataset_labels = ['label1', 'label2', 'label3']

    # create calculator frame
    calc_frame = CalculatorFrame(name="MyCalcFrame", datasets=[Data(data=data, dim_order='ps') for data in datasets], 
                                 names=dataset_names, labels=dataset_labels, subset='fabfour')
    assert(isinstance(calc_frame, CalculatorFrame)), "CalculatorFrame failed to instantiate."

    # check the properties of the frame
    # check expected number of calcs in frame - 3 for 3 datasets
    num_calcs_in_frame = calc_frame.n_calculators
    assert num_calcs_in_frame == 3, f"Unexpected number ({num_calcs_in_frame}) of calculators in the frame. Expected 3."
    
    # get the frame name
    assert calc_frame.name == "MyCalcFrame", "Calculator frame has unexpected name."

    # ensure dataset names, labels passed along to inidividual calculators
    for (index, calc) in enumerate(calc_frame.calculators[0]):
        assert calc.name == dataset_names[index], "Indiviudal calculator has unexpected name."
        assert calc.labels == dataset_labels[index], "Indiviudal calculator has unexpected label."
    
    # check that compute runs
    calc_frame.compute()

def test_correlation_frame_normal_operation():
    """Test whether the correlation frame instantiates as expected.""" 
    datasets = [np.random.randn(3, 100) for _ in range(3)]
    dataset_names = ['d1', 'd2', 'd3']
    dataset_labels = ['label1', 'label2', 'label3']
    calc_frame = CalculatorFrame(name="MyCalcFrame", datasets=[Data(data=data, dim_order='ps') for data in datasets], 
                                 names=dataset_names, labels=dataset_labels, subset='fabfour')
    
    calc_frame.compute()
    cf = calc_frame.get_correlation_df()

    assert not(cf[0].empty), "Correlation frame is empty."

def test_normalisation_flag():
    """Test whether the normalisation flag when instantiating
    the calculator works as expected."""
    data = np.random.randn(3, 100)
    calc = Calculator(dataset=data, normalise=False, detrend=False)
    calc_loaded_dataset = calc.dataset.to_numpy().squeeze()
    
    assert (calc_loaded_dataset == data).all(), f"Calculator normalise=False not producing the correct output." 
    


# For some reason the JVM causes a segfault with OpenBLAS (numpy's linalg sovler). Need to halt multithreading before starting JVM:
import os, logging, sys

os.environ['OMP_NUM_THREADS'] = '1'

# formatter = logging.Formatter('[%(levelname)s: %(asctime)s]: %(message)s')

# ch = logging.StreamHandler()
# ch.setFormatter(formatter)
# ch.setLevel(logging.DEBUG)

# logger = logging.getLogger()
# logger.addHandler(ch)
# logger.setLevel(logging.INFO)

# logging.captureWarnings(True)
# logging.basicConfig(stream=, level=logging.INFO)


import numpy as np
from sklearn.metrics import pairwise_distances
import tslearn.metrics
from tslearn.barycenters import (
    euclidean_barycenter,
    dtw_barycenter_averaging,
    dtw_barycenter_averaging_subgradient,
    softdtw_barycenter,
)
from hyppo.independence import (
    MGC,
    Dcorr,
    HHG,
    Hsic,
)
from hyppo.time_series import MGCX, DcorrX

from pyspi.base import (
    Directed,
    Undirected,
    Unsigned,
    Signed,
    parse_bivariate,
    parse_multivariate,
)


class PairwiseDistance(Undirected, Unsigned):

    name = "Pairwise distance"
    identifier = "pdist"
    labels = ["unsigned", "distance", "unordered", "nonlinear", "undirected"]

    def __init__(self, metric="euclidean", **kwargs):
        self._metric = metric
        self.identifier += f"_{metric}"

    @parse_multivariate
    def multivariate(self, data):
        return pairwise_distances(data.to_numpy(squeeze=True), metric=self._metric)


""" TODO: include optional kernels in each method
"""


class HilbertSchmidtIndependenceCriterion(Undirected, Unsigned):
    """Hilbert-Schmidt Independence Criterion (HSIC)"""

    name = "Hilbert-Schmidt Independence Criterion"
    identifier = "hsic"
    labels = ["unsigned", "distance", "unordered", "nonlinear", "undirected"]

    def __init__(self, biased=False):
        self._biased = biased
        if biased:
            self.identifier += "_biased"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        x, y = data.to_numpy()[[i, j]]
        stat = Hsic(bias=self._biased).statistic(x, y)
        return stat


class HellerHellerGorfine(Directed, Unsigned):
    """Heller-Heller-Gorfine independence criterion"""

    name = "Heller-Heller-Gorfine Independence Criterion"
    identifier = "hhg"
    labels = ["unsigned", "distance", "unordered", "nonlinear", "directed"]

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        x, y = data.to_numpy()[[i, j]]
        stat = HHG().statistic(x, y)
        return stat


class DistanceCorrelation(Undirected, Unsigned):
    """Distance correlation"""

    name = "Distance correlation"
    identifier = "dcorr"
    labels = ["unsigned", "distance", "unordered", "nonlinear", "undirected"]

    def __init__(self, biased=False):
        self._biased = biased
        if biased:
            self.identifier += "_biased"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        """ """
        x, y = data.to_numpy()[[i, j]]
        stat = Dcorr(bias=self._biased).statistic(x, y)
        return stat


class MultiscaleGraphCorrelation(Undirected, Unsigned):
    """Multiscale graph correlation"""

    name = "Multiscale graph correlation"
    identifier = "mgc"
    labels = ["distance", "unsigned", "unordered", "nonlinear", "undirected"]

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        x, y = data.to_numpy()[[i, j]]
        stat = MGC().statistic(x, y)
        return stat


class CrossDistanceCorrelation(Directed, Unsigned):
    """Cross-distance correlation"""

    name = "Cross-distance correlation"
    identifier = "dcorrx"
    labels = ["distance", "unsigned", "temporal", "directed", "nonlinear"]

    def __init__(self, max_lag=1):
        self._max_lag = max_lag
        self.identifier += f"_maxlag-{max_lag}"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy()
        x = z[i]
        y = z[j]
        stat, _ = DcorrX(max_lag=self._max_lag).statistic(x, y)
        return stat


class CrossMultiscaleGraphCorrelation(Directed, Unsigned):
    """Cross-multiscale graph correlation"""

    name = "Cross-multiscale graph correlation"
    identifier = "mgcx"
    labels = ["unsigned", "distance", "temporal", "directed", "nonlinear"]

    def __init__(self, max_lag=1):
        self._max_lag = max_lag
        self.identifier += f"_maxlag-{max_lag}"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy()
        x = z[i]
        y = z[j]
        stat, _, _ = MGCX(max_lag=self._max_lag).statistic(x, y)
        return stat


class TimeWarping(Undirected, Unsigned):

    labels = ["unsigned", "distance", "temporal", "undirected", "nonlinear"]

    def __init__(self, global_constraint=None):
        gcstr = global_constraint
        if gcstr is not None:
            gcstr = gcstr.replace("_", "-")
            self.identifier += f"_constraint-{gcstr}"
        self._global_constraint = global_constraint

    @property
    def simfn(self):
        try:
            return self._simfn
        except AttributeError:
            raise NotImplementedError(
                f"Add the similarity function for {self.identifier}"
            )

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy(squeeze=True)
        return self._simfn(z[i], z[j], global_constraint=self._global_constraint)


class DynamicTimeWarping(TimeWarping):

    name = "Dynamic time warping"
    identifier = "dtw"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._simfn = tslearn.metrics.dtw


class LongestCommonSubsequence(TimeWarping):

    name = "Longest common subsequence"
    identifier = "lcss"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._simfn = tslearn.metrics.lcss


class SoftDynamicTimeWarping(TimeWarping):

    name = "Dynamic time warping"
    identifier = "softdtw"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        z = data.to_numpy(squeeze=True)
        return tslearn.metrics.soft_dtw(z[i], z[j])


class Barycenter(Directed, Signed):

    name = "Barycenter"
    identifier = "bary"
    labels = ["distance", "signed", "undirected", "temporal", "nonlinear"]

    def __init__(self, mode="euclidean", squared=False, statistic="mean"):
        if mode == "euclidean":
            self._fn = euclidean_barycenter
        elif mode == "dtw":
            self._fn = dtw_barycenter_averaging
        elif mode == "sgddtw":
            self._fn = dtw_barycenter_averaging_subgradient
        elif mode == "softdtw":
            self._fn = softdtw_barycenter
        else:
            raise NameError(f"Unknown Barycenter mode: {mode}")
        self._mode = mode

        self._squared = squared
        self._preproc = lambda x: x
        if squared:
            self._preproc = lambda x: x**2
            self.identifier += f"-sq"

        if statistic == "mean":
            self._statfn = lambda x: np.nanmean(self._preproc(x))
        elif statistic == "max":
            self._statfn = lambda x: np.nanmax(self._preproc(x))
        else:
            raise NameError(f"Unknown statistic: {statistic}")

        self.identifier += f"_{mode}_{statistic}"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):

        try:
            bc = data.barycenter[self._mode][(i, j)]
        except (AttributeError, KeyError):
            z = data.to_numpy(squeeze=True)
            bc = self._fn(z[[i, j]])
            try:
                data.barycenter[self._mode][(i, j)] = bc
            except AttributeError:
                data.barycenter = {self._mode: {(i, j): bc}}
            except KeyError:
                data.barycenter[self._mode] = {(i, j): bc}
            data.barycenter[self._mode][(j, i)] = data.barycenter[self._mode][(i, j)]

        return self._statfn(bc)


class GromovWasserstainTau(Undirected, Unsigned):
    """Gromov-Wasserstain distance (GWTau)"""

    name = "Gromov-Wasserstain Distance"
    identifier = "gwtau"
    labels = ["unsigned", "distance", "unordered", "nonlinear", "undirected"]

    @staticmethod
    def vec_geo_dist(x):
        diffs = np.diff(x, axis=0)
        distances = np.linalg.norm(diffs, axis=1)
        return np.cumsum(distances)
    
    @staticmethod
    def wass_sorted(x1, x2):
        x1 = np.sort(x1)[::-1] # sort in descending order
        x2 = np.sort(x2)[::-1] 

        if len(x1) == len(x2):
            res = np.sqrt(np.mean((x1 - x2) ** 2))
        else:
            N, M = len(x1), len(x2)
            i_ratios = np.arange(1, N + 1) / N
            j_ratios = np.arange(1, M + 1) / M
        
        
            min_values = np.minimum.outer(i_ratios, j_ratios)
            max_values = np.maximum.outer(i_ratios - 1/N, j_ratios - 1/M)
        
            lam = np.where(min_values > max_values, min_values - max_values, 0)
        
            diffs_squared = (x1[:, None] - x2) ** 2
            my_sum = np.sum(lam * diffs_squared)
        
            res = np.sqrt(my_sum)

        return res
    
    @staticmethod
    def gwtau(xi, xj):
        timei = np.arange(len(xi))
        timej = np.arange(len(xj))
        traji = np.column_stack([timei, xi])
        trajj = np.column_stack([timej, xj])

        vi = GromovWasserstainTau.vec_geo_dist(traji)
        vj = GromovWasserstainTau.vec_geo_dist(trajj)
        gw = GromovWasserstainTau.wass_sorted(vi, vj)
    
        return gw

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        x, y = data.to_numpy()[[i, j]]
        # insert compute SPI code here (computes on x and y)
        stat = self.gwtau(x, y)
        return stat



import mne.connectivity as mnec
from pyspi.base import (
    Directed,
    Undirected,
    Unsigned,
    parse_bivariate,
    parse_multivariate,
)
import numpy as np
import warnings
from functools import partial


class mne(Unsigned):
    def __init__(self, fs=1, fmin=0, fmax=None, statistic="mean"):
        if fmax is None:
            fmax = fs / 2

        self._fs = fs
        if fs != 1:
            warnings.warn("Multiple sampling frequencies not yet handled.")
        self._fmin = fmin
        self._fmax = fmax
        if statistic == "mean":
            self._statfn = np.nanmean
        elif statistic == "max":
            self._statfn = np.nanmax
        else:
            raise NameError(f"Unknown statistic {statistic}")

        self._statistic = statistic

        paramstr = (
            f"_wavelet_{statistic}_fs-{fs}_fmin-{fmin:.3g}_fmax-{fmax:.3g}".replace(
                ".", "-"
            )
        )
        self.identifier += paramstr

    @property
    def measure(self):
        try:
            return self._measure
        except AttributeError:
            raise AttributeError(f"Include measure for {self.identifier}")

    def _get_cache(self, data):
        try:
            conn, freq = data.mne[self.measure]
        except (KeyError, AttributeError):
            z = np.moveaxis(data.to_numpy(), 2, 0)

            cwt_freqs = np.linspace(0.2, 0.5, 125)
            cwt_n_cycles = cwt_freqs / 7.0
            conn, freq, _, _, _ = mnec.spectral_connectivity(
                data=z,
                method=self.measure,
                mode="cwt_morlet",
                sfreq=self._fs,
                mt_adaptive=True,
                fmin=5 / data.n_observations,
                fmax=self._fs / 2,
                cwt_freqs=cwt_freqs,
                cwt_n_cycles=cwt_n_cycles,
                verbose="WARNING",
            )

            try:
                data.mne[self.measure] = (conn, freq)
            except AttributeError:
                data.mne = {self.measure: (conn, freq)}

        freq_id = np.where((freq >= self._fmin) * (freq <= self._fmax))[0]

        return conn, freq_id

    @parse_multivariate
    def multivariate(self, data):
        adj_freq, freq_id = self._get_cache(data)
        try:
            adj = self._statfn(adj_freq[..., freq_id, :], axis=(2, 3))
        except np.AxisError:
            adj = self._statfn(adj_freq[..., freq_id], axis=2)
        ui = np.triu_indices(data.n_processes, 1)
        adj[ui] = adj.T[ui]
        np.fill_diagonal(adj, np.nan)
        return adj


def modify_stats(statfn, modifier):
    def parsed_stats(stats, statfn, modifier, **kwargs):
        return statfn(modifier(stats), **kwargs)

    return partial(parsed_stats, statfn=statfn, modifier=modifier)


class CoherenceMagnitude(mne, Undirected):
    name = "Coherence magnitude (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "cohmag"
        self._measure = "coh"
        super().__init__(**kwargs)


class CoherencePhase(mne, Undirected):
    name = "Coherence phase (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "phase"
        self._measure = "cohy"
        super().__init__(**kwargs)

        # Take the angle before computing the statistic
        self._statfn = modify_stats(self._statfn, np.angle)


class ImaginaryCoherence(mne, Undirected):
    name = "Imaginary coherency (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "icoh"
        self._measure = "imcoh"
        super().__init__(**kwargs)


class PhaseLockingValue(mne, Undirected):
    name = "Phase locking value (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "plv"
        self._measure = "plv"
        super().__init__(**kwargs)


class PairwisePhaseConsistency(mne, Undirected):
    name = "Pairwise phase consistency (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "ppc"
        self._measure = "ppc"
        super().__init__(**kwargs)


class PhaseLagIndex(mne, Undirected):
    name = "Phase lag index (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "pli"
        self._measure = "pli"
        super().__init__(**kwargs)


class DebiasedSquaredWeightedPhaseLagIndex(mne, Undirected):
    name = "Debiased squared phase lag index (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "dspli"
        self._measure = "pli2_unbiased"
        super().__init__(**kwargs)


class weighted_PhaseLagIndex(mne, Undirected):
    name = "Weighted squared phase lag index (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "wspli"
        self._measure = "wpli"
        super().__init__(**kwargs)


class debiased_weighted_squared_PhaseLagIndex(mne, Undirected):
    name = "Debiased weighted squared phase lag index (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "dwspli"
        self._measure = "wpli2_debiased"
        super().__init__(**kwargs)


class PhaseSlopeIndex(mne, Undirected):
    name = "Phase slope index (wavelet)"
    labels = ["unsigned", "wavelet", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "psi"
        super().__init__(**kwargs)
        self.identifier += f"_{self._statistic}"

    def _get_cache(self, data):
        try:
            psi = data.mne_psi["psi"]
            freq = data.mne_psi["freq"]
        except AttributeError:
            z = np.moveaxis(data.to_numpy(), 2, 0)

            freqs = np.linspace(0.2, 0.5, 10)
            psi, freq, _, _, _ = mnec.phase_slope_index(
                data=z,
                mode="cwt_morlet",
                sfreq=self._fs,
                mt_adaptive=True,
                cwt_freqs=freqs,
                verbose="WARNING",
            )
            freq = freq[0]
            data.mne_psi = dict(psi=psi, freq=freq)

        # freq = conn.frequencies
        freq_id = np.where((freq >= self._fmin) * (freq <= self._fmax))[0]

        return psi, freq_id

    @parse_multivariate
    def multivariate(self, data):
        adj_freq, freq_id = self._get_cache(data)
        adj = self._statfn(np.real(adj_freq[..., freq_id]), axis=(2, 3))

        ui = np.triu_indices(data.n_processes, 1)
        adj[ui] = adj.T[ui]
        np.fill_diagonal(adj, np.nan)
        return adj



from pyspi.calculator import Calculator
import pytest
import dill
import pyspi
import numpy as np

############# Fixtures and helper functions #########

def load_benchmark_tables():
    """Function to load the mean and standard deviation tables for each MPI."""
    table_fname = 'CML7_benchmark_tables.pkl'
    with open(f"tests/{table_fname}", "rb") as f:
        loaded_tables = dill.load(f)
    
    return loaded_tables

def load_benchmark_dataset():
    dataset_fname = 'cml7.npy'
    dataset = np.load(f"pyspi/data/{dataset_fname}").T
    return dataset

def compute_new_tables():
    """Compute new tables using the same benchmark dataset(s)."""
    benchmark_dataset = load_benchmark_dataset()
    # Compute new tables on the benchmark dataset
    np.random.seed(42)
    calc = Calculator(dataset=benchmark_dataset, normalise=True, detrend=True)
    calc.compute()
    table_dict = dict()
    for spi in calc.spis:
        table_dict[spi] = calc.table[spi]

    return table_dict

def generate_SPI_test_params():
    """Function to generate combinations of benchmark table, 
    new table for each MPI"""
    benchmark_tables = load_benchmark_tables()
    new_tables = compute_new_tables()
    params = []
    calc = Calculator()
    spis = list(calc.spis.keys())
    spi_ob = list(calc.spis.values())
    for spi_est, spi_ob in zip(spis, spi_ob):
        params.append((spi_est, spi_ob, benchmark_tables[spi_est], new_tables[spi_est].to_numpy()))
    
    return params

params = generate_SPI_test_params()
def pytest_generate_tests(metafunc):
    """Create a hook to generate parameter combinations for parameterised test"""
    if "est" in metafunc.fixturenames:
        metafunc.parametrize("est, est_ob, mpi_benchmark,mpi_new", params)
        

def test_mpi(est, est_ob, mpi_benchmark, mpi_new, spi_warning_logger):
    """Run the benchmarking tests."""
    zscore_threshold = 1 # 2 sigma
    
    # separate the the mean and std. dev tables for the benchmark
    mean_table = mpi_benchmark['mean']
    std_table = mpi_benchmark['std']

    # check std table for zeros and impute with smallest non-zero value
    std_table = np.where(std_table == 0, 1e-10, std_table)
    
    # check that the shapes are equal
    assert mean_table.shape == mpi_new.shape, f"SPI: {est}| Different table shapes. "

    # convert NaNs to zeros before proeceeding - this will take care of diagonal and any null outputs
    mpi_new = np.nan_to_num(mpi_new)
    mpi_mean = np.nan_to_num(mean_table)

    # check if matrix is symmetric (undirected SPI) for num exceed correction
    isSymmetric = "undirected" in est_ob.labels 

    # get the module name for easy reference
    module_name = est_ob.__module__.split(".")[-1]

    if not np.allclose(mpi_new, mpi_mean):
        # tables are not equivalent, quantify the difference by z-scoring.
        diff = abs(mpi_new - mpi_mean)
        zscores = diff/std_table

        idxs_greater_than_thresh = np.argwhere(zscores > zscore_threshold)

        if len(idxs_greater_than_thresh) > 0:
            sigs = zscores[idxs_greater_than_thresh[:, 0], idxs_greater_than_thresh[:, 1]]
            # get the max
            max_z = max(sigs)

            # number of interactions
            num_interactions = mpi_new.size - mpi_new.shape[0]
            # count exceedances
            num_exceed = len(sigs)

            if isSymmetric:
                # number of unique exceedences is half
                num_exceed //= 2
                num_interactions //= 2

            spi_warning_logger(est, module_name, max_z, int(num_exceed), int(num_interactions))



import jpype as jp
import numpy as np
from pyspi import utils
try:
    from oct2py import octave, Struct
except Exception:
    pass
import copy
import os
import logging

from pyspi.base import Undirected, Directed, Unsigned, parse_univariate, parse_bivariate

"""
Contains relevant dependence statistics from the information theory community.
"""
# if not jp.isJVMStarted():
#     jarloc = (
#         os.path.dirname(os.path.abspath(__file__)) + "/../lib/jidt/infodynamics.jar"
#     )
#     # Change to debug info
#     logging.debug(f"Starting JVM with java class {jarloc}.")
#     jp.startJVM(jp.getDefaultJVMPath(), "-ea", "-Djava.class.path=" + jarloc)


class JIDTBase(Unsigned):

    # List of (currently) modifiable parameters
    _NNK_PROP_NAME = "k"
    _AUTO_EMBED_METHOD_PROP_NAME = "AUTO_EMBED_METHOD"
    _DYN_CORR_EXCL_PROP_NAME = "DYN_CORR_EXCL"
    _KERNEL_WIDTH_PROP_NAME = "KERNEL_WIDTH"
    _K_HISTORY_PROP_NAME = "k_HISTORY"
    _K_TAU_PROP_NAME = "k_TAU"
    _L_HISTORY_PROP_NAME = "l_HISTORY"
    _L_TAU_PROP_NAME = "l_TAU"
    _K_SEARCH_MAX_PROP_NAME = "AUTO_EMBED_K_SEARCH_MAX"
    _TAU_SEARCH_MAX_PROP_NAME = "AUTO_EMBED_TAU_SEARCH_MAX"
    _BIAS_CORRECTION = "BIAS_CORRECTION"
    _NORMALISE = "NORMALISE"
    _SEED = "NOISE_SEED"

    _base_class = jp.JPackage("infodynamics.measures.continuous")

    def __init__(
        self, estimator="gaussian", kernel_width=0.5, prop_k=4, dyn_corr_excl=None
    ):

        self._estimator = estimator
        self._kernel_width = kernel_width
        self._prop_k = prop_k
        self._dyn_corr_excl = dyn_corr_excl
        self._entropy_calc = self._getcalc("entropy")

        self.identifier = self.identifier + "_" + estimator
        if estimator == "kraskov":
            self.identifier = self.identifier + "_NN-{}".format(prop_k)
            self.labels = self.labels + ["nonlinear"]
        elif estimator == "kernel":
            self.identifier = self.identifier + "_W-{}".format(kernel_width)
            self.labels = self.labels + ["nonlinear"]
        elif estimator == "symbolic":
            if not isinstance(self, TransferEntropy):
                raise NotImplementedError(
                    "Symbolic estimator is only available for transfer entropy."
                )
            self.labels = self.labels + ["symbolic"]
            self._dyn_corr_excl = None
            return
        else:
            self.labels = self.labels + ["linear"]
            self._dyn_corr_excl = None

        if self._dyn_corr_excl:
            self.identifier = self.identifier + "_DCE"

    def __getstate__(self):
        state = dict(self.__dict__)

        unserializable_objects = ["_entropy_calc", "_calc"]

        for k in unserializable_objects:
            if k in state.keys():
                del state[k]

        return state

    def __setstate__(self, state):
        """Re-initialise the calculator"""
        # Re-initialise
        self.__dict__.update(state)
        self._entropy_calc = self._getcalc("entropy")

    def __deepcopy__(self, memo):
        newone = type(self)()
        newone.__dict__.update(self.__dict__)
        for attr in newone.__dict__:
            setattr(newone, attr, copy.deepcopy(getattr(self, attr), memo))
        return newone

    def _setup(self, calc):
        if self._estimator == "kernel":
            calc.setProperty(self._KERNEL_WIDTH_PROP_NAME, str(self._kernel_width))
        elif self._estimator == "kraskov":
            calc.setProperty(self._NNK_PROP_NAME, str(self._prop_k))

        calc.setProperty(self._BIAS_CORRECTION, "false")
        calc.setProperty(self._SEED, "42")

        return calc

    def _getkey(self):
        if self._estimator == "kernel":
            return (self._estimator, self._kernel_width)
        elif self._estimator == "kraskov":
            return (self._estimator, self._prop_k)
        else:
            return (self._estimator,)

    def _getcalc(self, measure):
        if measure == "entropy":
            if self._estimator == "kernel":
                calc = self._base_class.kernel.EntropyCalculatorMultiVariateKernel()
            elif self._estimator == "kozachenko":
                calc = (
                    self._base_class.kozachenko.EntropyCalculatorMultiVariateKozachenko()
                )
            else:
                calc = self._base_class.gaussian.EntropyCalculatorMultiVariateGaussian()
        elif measure == "MutualInfo":
            if self._estimator == "kernel":
                calc = self._base_class.kernel.MutualInfoCalculatorMultiVariateKernel()
            elif self._estimator == "kraskov":
                calc = (
                    self._base_class.kraskov.MutualInfoCalculatorMultiVariateKraskov1()
                )
            else:
                calc = (
                    self._base_class.gaussian.MutualInfoCalculatorMultiVariateGaussian()
                )
        elif measure == "TransferEntropy":
            if self._estimator == "kernel":
                calc = self._base_class.kernel.TransferEntropyCalculatorKernel()
            elif self._estimator == "kraskov":
                calc = self._base_class.kraskov.TransferEntropyCalculatorKraskov()
            else:
                calc = self._base_class.gaussian.TransferEntropyCalculatorGaussian()
        else:
            raise TypeError(f"Unknown measure: {measure}")

        return self._setup(calc)

    # No Theiler window yet (can it be done?)
    @parse_univariate
    def _compute_entropy(self, data, i=None):
        if not hasattr(data, "entropy"):
            data.entropy = {}

        key = self._getkey()
        if key not in data.entropy:
            data.entropy[key] = np.full((data.n_processes,), -np.inf)

        if data.entropy[key][i] == -np.inf:
            x = np.squeeze(data.to_numpy()[i])

            self._entropy_calc.initialise(1)
            self._entropy_calc.setObservations(jp.JArray(jp.JDouble, 1)(x))

            data.entropy[key][
                i
            ] = self._entropy_calc.computeAverageLocalOfObservations()

        return data.entropy[key][i]

    # No Theiler window is available in the JIDT estimator
    @parse_bivariate
    def _compute_joint_entropy(self, data, i, j):
        if not hasattr(data, "joint_entropy"):
            data.joint_entropy = {}

        key = self._getkey()
        if key not in data.joint_entropy:
            data.joint_entropy[key] = np.full((data.n_processes, data.n_processes), -np.infty)

        if data.joint_entropy[key][i, j] == -np.inf:
            x, y = data.to_numpy()[[i, j]]

            self._entropy_calc.initialise(2)
            self._entropy_calc.setObservations(jp.JArray(jp.JDouble, 2)(np.concatenate([x, y], axis=1)))

            data.joint_entropy[key][i, j] = self._entropy_calc.computeAverageLocalOfObservations()
            data.joint_entropy[key][j, i] = data.joint_entropy[key][i, j]

        return data.joint_entropy[key][i, j]

    # No Theiler window is available in the JIDT estimator
    def _compute_conditional_entropy(self, X, Y):
        XY = np.concatenate([X, Y], axis=1)

        self._entropy_calc.initialise(XY.shape[1])
        self._entropy_calc.setObservations(jp.JArray(jp.JDouble, XY.ndim)(XY))

        H_XY = self._entropy_calc.computeAverageLocalOfObservations()

        self._entropy_calc.initialise(Y.shape[1])
        self._entropy_calc.setObservations(jp.JArray(jp.JDouble, Y.ndim)(Y))

        H_Y = self._entropy_calc.computeAverageLocalOfObservations()

        return H_XY - H_Y

    def _set_theiler_window(self, data, i, j):
        if self._dyn_corr_excl == "AUTO":
            if not hasattr(data, "theiler"):
                z = data.to_numpy()
                theiler_window = -np.ones((data.n_processes, data.n_processes))

                # Compute effective sample size for each pair
                for _i in range(data.n_processes):
                    targ = z[_i]
                    for _j in range(_i + 1, data.n_processes):
                        src = z[_j]

                        # Initialize the Theiler window using Bartlett's formula
                        theiler_window[_i, _j] = 2 * np.dot(
                            utils.acf(src), utils.acf(targ)
                        )
                        theiler_window[_j, _i] = theiler_window[_i, _j]
                data.theiler = theiler_window

            self._calc.setProperty(
                self._DYN_CORR_EXCL_PROP_NAME, str(int(data.theiler[i, j]))
            )
        elif self._dyn_corr_excl is not None:
            self._calc.setProperty(
                self._DYN_CORR_EXCL_PROP_NAME, str(int(self._dyn_corr_excl))
            )


class JointEntropy(JIDTBase, Undirected):

    name = "Joint entropy"
    identifier = "je"
    labels = ["unsigned", "infotheory", "unordered", "undirected"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        return self._compute_joint_entropy(data, i=i, j=j)


class ConditionalEntropy(JIDTBase, Directed):

    name = "Conditional entropy"
    identifier = "ce"
    labels = ["unsigned", "infotheory", "unordered", "directed"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        return self._compute_joint_entropy(data, i=i, j=j) - self._compute_entropy(
            data, i=i
        )


class MutualInfo(JIDTBase, Undirected):
    name = "Mutual information"
    identifier = "mi"
    labels = ["unsigned", "infotheory", "unordered", "undirected"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._calc = self._getcalc("MutualInfo")

    def __setstate__(self, state):
        super().__setstate__(state)
        self.__dict__.update(state)
        self._calc = self._getcalc("MutualInfo")

    @parse_bivariate
    def bivariate(self, data, i=None, j=None, verbose=False):
        """Compute mutual information between Y and X"""
        self._set_theiler_window(data, i, j)
        self._calc.initialise(1, 1)

        try:
            src, targ = data.to_numpy(squeeze=True)[[i, j]]
            self._calc.setObservations(
                jp.JArray(jp.JDouble)(src), jp.JArray(jp.JDouble)(targ)
            )
            return self._calc.computeAverageLocalOfObservations()
        except:
            logging.warning(
                "MI calcs failed. Maybe check input data for Cholesky factorisation?"
            )
            return np.nan


class TimeLaggedMutualInfo(JIDTBase, Directed):
    name = "Time-lagged mutual information"
    identifier = "tlmi"
    labels = ["unsigned", "infotheory", "temporal", "directed"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._calc = self._getcalc("MutualInfo")

    def __setstate__(self, state):
        """Re-initialise the calculator"""
        super().__setstate__(state)
        self.__dict__.update(state)
        self._calc = self._getcalc("MutualInfo")

    @parse_bivariate
    def bivariate(self, data, i=None, j=None, verbose=False):
        self._set_theiler_window(data, i, j)
        self._calc.initialise(1, 1)
        try:
            src, targ = data.to_numpy(squeeze=True)[[i, j]]
            src = src[:-1]
            targ = targ[1:]
            self._calc.setObservations(
                jp.JArray(jp.JDouble, 1)(src), jp.JArray(jp.JDouble, 1)(targ)
            )
            return self._calc.computeAverageLocalOfObservations()
        except:
            logging.warning(
                "Time-lagged MI calcs failed. Maybe check input data for Cholesky factorisation?"
            )
            return np.nan


class TransferEntropy(JIDTBase, Directed):

    name = "Transfer entropy"
    identifier = "te"
    labels = ["unsigned", "embedding", "infotheory", "temporal", "directed"]

    def __init__(
        self,
        auto_embed_method=None,
        k_search_max=None,
        tau_search_max=None,
        k_history=1,
        k_tau=1,
        l_history=1,
        l_tau=1,
        **kwargs,
    ):

        if "estimator" not in kwargs.keys() or kwargs["estimator"] == "gaussian":
            self.identifier = "gc"
        super().__init__(**kwargs)
        self._calc = self._getcalc("TransferEntropy")

        # Auto-embedding
        if auto_embed_method is not None:
            self._calc.setProperty(self._AUTO_EMBED_METHOD_PROP_NAME, auto_embed_method)
            self._calc.setProperty(self._K_SEARCH_MAX_PROP_NAME, str(k_search_max))
            if self._estimator != "kernel":
                self.identifier = self.identifier + "_k-max-{}_tau-max-{}".format(
                    k_search_max, tau_search_max
                )
                self._calc.setProperty(
                    self._TAU_SEARCH_MAX_PROP_NAME, str(tau_search_max)
                )
            else:
                self.identifier = self.identifier + "_k-max-{}".format(k_search_max)
            # Set up calculator
        else:
            self._calc.setProperty(self._K_HISTORY_PROP_NAME, str(k_history))
            if self._estimator != "kernel":
                self._calc.setProperty(self._K_TAU_PROP_NAME, str(k_tau))
                self._calc.setProperty(self._L_HISTORY_PROP_NAME, str(l_history))
                self._calc.setProperty(self._L_TAU_PROP_NAME, str(l_tau))
                self.identifier = self.identifier + "_k-{}_kt-{}_l-{}_lt-{}".format(
                    k_history, k_tau, l_history, l_tau
                )
            else:
                self.identifier = self.identifier + "_k-{}".format(k_history)

    def __setstate__(self, state):
        """Re-initialise the calculator"""
        # Re-initialise
        super().__setstate__(state)
        self.__dict__.update(state)
        self._calc = self._getcalc("TransferEntropy")

    @parse_bivariate
    def bivariate(self, data, i=None, j=None, verbose=False):
        """
        Compute transfer entropy from i->j
        """
        self._set_theiler_window(data, i, j)
        self._calc.initialise()
        src, targ = data.to_numpy(squeeze=True)[[i, j]]
        try:
            self._calc.setObservations(
                jp.JArray(jp.JDouble, 1)(src), jp.JArray(jp.JDouble, 1)(targ)
            )
            return self._calc.computeAverageLocalOfObservations()
        except Exception as err:
            logging.warning(f"TE calcs failed: {err}.")
            return np.nan


class CrossmapEntropy(JIDTBase, Directed):

    name = "Cross-map entropy"
    identifier = "xme"
    labels = ["unsigned", "infotheory", "temporal", "directed"]

    def __init__(self, history_length=10, **kwargs):
        super().__init__(**kwargs)
        self.identifier += f"_k{history_length}"
        self._history_length = history_length

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        src, targ = data.to_numpy(squeeze=True)[[i, j]]
        k = self._history_length
        targ_future = targ[k:]
        src_past = np.expand_dims(src[k - 1 : -1], axis=1)
        for i in range(2, k):
            src_past = np.append(
                src_past, np.expand_dims(src[k - i : -i], axis=1), axis=1
            )

        joint = np.concatenate([src_past, np.expand_dims(targ_future, axis=1)], axis=1)

        self._entropy_calc.initialise(joint.shape[1])
        self._entropy_calc.setObservations(jp.JArray(jp.JDouble, 2)(joint))
        H_xy = self._entropy_calc.computeAverageLocalOfObservations()

        self._entropy_calc.initialise(src_past.shape[1])
        self._entropy_calc.setObservations(jp.JArray(jp.JDouble, 2)(src_past))
        H_y = self._entropy_calc.computeAverageLocalOfObservations()

        return H_xy - H_y


class CausalEntropy(JIDTBase, Directed):

    name = "Causally conditioned entropy"
    identifier = "cce"
    labels = ["unsigned", "infotheory", "temporal", "directed"]

    def __init__(self, n=5, **kwargs):
        super().__init__(**kwargs)
        self._n = n

    def _compute_causal_entropy(self, src, targ):

        src = np.squeeze(src)
        targ = np.squeeze(targ)

        m_utils = jp.JPackage("infodynamics.utils").MatrixUtils

        causal_entropy = 0
        for i in range(1, self._n + 1):
            Yp = m_utils.makeDelayEmbeddingVector(jp.JArray(jp.JDouble, 1)(targ), i - 1)[:-1]
            Xp = m_utils.makeDelayEmbeddingVector(jp.JArray(jp.JDouble, 1)(src), i)
            XYp = np.concatenate([Yp, Xp], axis=1)

            Yf = np.expand_dims(targ[i - 1 :], 1)
            causal_entropy = causal_entropy + self._compute_conditional_entropy(Yf, XYp)
        return causal_entropy

    def _getkey(self):
        return super(CausalEntropy, self)._getkey() + (self._n,)

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        if not hasattr(data, "causal_entropy"):
            data.causal_entropy = {}

        key = self._getkey()
        if key not in data.causal_entropy:
            data.causal_entropy[key] = np.full(
                (data.n_processes, data.n_processes), -np.inf
            )

        if data.causal_entropy[key][i, j] == -np.inf:
            z = data.to_numpy(squeeze=True)
            data.causal_entropy[key][i, j] = self._compute_causal_entropy(z[i], z[j])

        return data.causal_entropy[key][i, j]


class DirectedInfo(CausalEntropy, Directed):

    name = "Directed information"
    identifier = "di"
    labels = ["unsigned", "infotheory", "temporal", "directed"]

    def __init__(self, n=5, **kwargs):
        super().__init__(**kwargs)
        self._n = n

    def _compute_entropy_rates(self, targ):

        targ = np.squeeze(targ)
        m_utils = jp.JPackage("infodynamics.utils").MatrixUtils

        entropy_rate_sum = 0
        for i in range(1, self._n + 1):
            # Compute entropy for an i-dimensional embedding
            self._entropy_calc.initialise(i)

            Yi = m_utils.makeDelayEmbeddingVector(jp.JArray(jp.JDouble, 1)(targ), i)
            self._entropy_calc.setObservations(Yi)
            entropy_rate_sum = entropy_rate_sum + self._entropy_calc.computeAverageLocalOfObservations() / i

        return entropy_rate_sum

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        """Compute directed information from i to j"""

        entropy_rates = self._compute_entropy_rates(data.to_numpy(squeeze=True)[j])
        causal_entropy = super().bivariate(data, i=i, j=j)

        return entropy_rates - causal_entropy


class StochasticInteraction(JIDTBase, Undirected):

    name = "Stochastic interaction"
    identifier = "si"
    labels = ["unsigned", "infotheory", "temporal", "undirected"]

    def __init__(self, delay=1, **kwargs):
        super().__init__(**kwargs)
        self._delay = delay
        self.identifier += f"_k-{delay}"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None, verbose=False):
        x, y = data.to_numpy()[[i, j]]
        xy = np.concatenate([x, y], axis=1)
        tau = self._delay

        H_joint = self._compute_conditional_entropy(xy[tau:], xy[:-tau])
        H_src = self._compute_conditional_entropy(x[tau:], x[:-tau])
        H_targ = self._compute_conditional_entropy(y[tau:], y[:-tau])

        return H_src + H_targ - H_joint


class IntegratedInformation(Undirected, Unsigned):

    name = "Integrated information"
    identifier = "phi"
    labels = ["linear", "unsigned", "infotheory", "temporal", "undirected"]

    def __init__(self, phitype="star", delay=1, normalization=0):
        self._params = Struct()
        self._params["tau"] = 1
        self._options = Struct()
        self._options["type_of_phi"] = phitype
        self._options["type_of_dist"] = "Gauss"
        self._options["normalization"] = normalization
        self.identifier += f"_{phitype}_t-{delay}_norm-{normalization}"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):

        if not octave.exist("phi_comp"):
            path = os.path.dirname(os.path.abspath(__file__)) + "/../lib/PhiToolbox/"
            octave.addpath(octave.genpath(path))

        P = [1, 2]
        Z = data.to_numpy(squeeze=True)[[i, j]]

        return octave.phi_comp(Z, P, self._params, self._options)



from setuptools import setup, find_packages

# http://www.diveintopython3.net/packaging.html
# https://pypi.python.org/pypi?:action=list_classifiers

with open('README.md', 'r', encoding='utf-8') as file:
    long_description = file.read()


install_requires = [
        'h5py',
        'scikit-learn',
        'scipy',
        'numpy<2.0.0',
        'pandas',
        'statsmodels',
        'pyyaml',
        'tqdm',
        'nitime',
        'hyppo',
        'pyEDM==1.15.2.0',
        'jpype1',
        'sktime',
        'dill',
        'spectral-connectivity',
        'torch',
        'cdt==0.5.23',
        'oct2py',
        'tslearn',
        'mne==0.23.0',
        'seaborn',
        'future',
        'colorama'
]

testing_extras = [
    'pytest==5.4.2',  # unittest.TestCase funkyness, see commit 77c1505ab
]


setup(
    name='pyspi',
    packages=find_packages(),
    package_data={'': ['config.yaml',
                       'sonnet_config.yaml',
                        'fast_config.yaml',
                        'fabfour_config.yaml',
                        'lib/jidt/infodynamics.jar',
                        'lib/PhiToolbox/Phi/phi_comp.m',
                        'lib/PhiToolbox/Phi/phi_star_Gauss.m',
                        'lib/PhiToolbox/Phi/phi_G_Gauss.m',
                        'lib/PhiToolbox/Phi/phi_G_Gauss_AL.m',
                        'lib/PhiToolbox/Phi/phi_G_Gauss_LBFGS.m',
                        'lib/PhiToolbox/Phi/phi_comp_probs.m',
                        'lib/PhiToolbox/Phi/phi_Gauss.m',
                        'lib/PhiToolbox/utility/I_s_I_s_d.m',
                        'lib/PhiToolbox/utility/data_to_probs.m',
                        'lib/PhiToolbox/utility/Gauss/Cov_comp.m',
                        'lib/PhiToolbox/utility/Gauss/Cov_cond.m',
                        'lib/PhiToolbox/utility/Gauss/H_gauss.m',
                        'lib/PhiToolbox/utility/Gauss/logdet.m',
                        'data/cml.npy',
                        'data/forex.npy',
                        'data/standard_normal.npy',
                        'data/cml7.npy']},
    include_package_data=True,
    version='2.0.0',
    description='Library for pairwise analysis of time series data.',
    author='Oliver M. Cliff',
    author_email='oliver.m.cliff@gmail.com',
    url='https://github.com/DynamicsAndNeuralSystems/pyspi',
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 1 - Planning",
        "Operating System :: POSIX :: Linux",
        "Intended Audience :: Science/Research",
        "Environment :: Console",
        "Environment :: Other Environment",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
    ],
    install_requires=install_requires,
    extras_require={'testing': testing_extras}
)



"""Provide data structures for multivariate analysis.

Code is adapted from Patricia Wollstadt's IDTxL (https://github.com/pwollstadt/IDTxl)
"""
import numpy as np
import pandas as pd
from pyspi import utils
from scipy.stats import zscore
from scipy.signal import detrend
from colorama import init, Fore
import os

VERBOSE = False
init(autoreset=True) # automatically reset coloured outputs

class Data:
    """Store data for dependency analysis.

    Data takes a 2-dimensional array representing realisations of random
    variables in dimensions: processes and observations.
    Indicate the actual order of dimensions in the provided array in a two-character string, e.g. 'ps'
    for an array with realisations over (1) processes, and (2) observations in time.

    Example:
        >>> # Initialise empty data object
        >>> data = Data()
        >>>
        >>> # Load a prefilled financial dataset
        >>> data_forex = Data().load_dataset(forex)
        >>>
        >>> # Create data objects with data of various sizes
        >>> d = np.arange(3000).reshape((3, 1000))  # 3 procs.,
        >>> data_2 = Data(d, dim_order='ps')        # 1000 observations
        >>>
        >>> # Overwrite data in existing object with random data
        >>> d = np.arange(5000)
        >>> data_2.set_data(data_new, 's')

    Args:
        data (array_like, optional):
            2-dimensional array with raw data, defaults to None.
        dim_order (str, optional):
            Order of dimensions, accepts two combinations of the characters 'p', and 's' for processes and observations, defaults to 'ps'.
        detrend (bool, optional):
            If True, detrend the dataset along the time axis before normalising (if enabled), defaults to True.
        normalise (bool, optional):
            If True, z-score normalise the dataset along the time axis before computing SPIs, defaults to True.
            Detrending (if enabled) is always applied before normalisation.
        name (str, optional):
            Name of the dataset
        procnames (list, optional):
            List of process names with length the number of processes, defaults to None.
        n_processes (int, optional):
            Truncates data to this many processes, defaults to None.
        n_observations (int, optional):
            Truncates data to this many observations, defaults to None.

    """

    def __init__(
        self,
        data=None,
        dim_order="ps",
        detrend=True,
        normalise=True,
        name=None,
        procnames=None,
        n_processes=None,
        n_observations=None,
    ):
        self.normalise = normalise
        self.detrend = detrend
        if data is not None:
            dat = self.convert_to_numpy(data)
            self.set_data(
                dat,
                dim_order=dim_order,
                name=name,
                n_processes=n_processes,
                n_observations=n_observations,
            )

        if procnames is not None:
            assert len(procnames) == self.n_processes

    @property
    def name(self):
        """Name of the data object."""
        if hasattr(self, "_name"):
            return self._name
        else:
            return "N/A"

    @name.setter
    def name(self, n):
        """Set the name of the data object."""
        if not isinstance(n, str):
            raise TypeError(f"Name should be a string, received {type(n)}.")
        self._name = n

    @property
    def procnames(self):
        """List of process names."""
        if hasattr(self, "_procnames"):
            return self._procnames
        else:
            return [f"proc-{i}" for i in range(self.n_processes)]

    def to_numpy(self, realisation=None, squeeze=False):
        """Return the numpy array."""
        if realisation is not None:
            dat = self._data[:, :, realisation]
        else:
            dat = self._data

        if squeeze:
            return np.squeeze(dat)
        else:
            return dat

    @staticmethod
    def convert_to_numpy(data):
        """Converts other data instances to default numpy format."""

        if isinstance(data, np.ndarray):
            npdat = data
        elif isinstance(data, pd.DataFrame):
            npdat = data.to_numpy()
        elif isinstance(data, str):
            ext = os.path.splitext(data)[1]
            if ext == ".npy":
                npdat = np.load(data)
            elif ext == ".txt":
                npdat = np.genfromtxt(data)
            elif ext == ".csv":
                npdat = np.genfromtxt(data, ",")
            elif ext == ".ts":
                from sktime.utils.data_io import load_from_tsfile_to_dataframe
                from sktime.datatypes._panel._convert import from_nested_to_3d_numpy

                tsdat, _ = load_from_tsfile_to_dataframe(data)
                npdat = from_nested_to_3d_numpy(tsdat)
            else:
                raise TypeError(f"Unknown filename extension: {ext}")
        else:
            raise TypeError(f"Unknown data type: {type(data)}")

        return npdat

    def set_data(
        self,
        data,
        dim_order="ps",
        name=None,
        n_processes=None,
        n_observations=None,
        verbose=False,
    ):
        """Overwrite data in an existing instance.

        Args:
            data (np.ndarray):
                2-dimensional array of realisations
            dim_order (str, optional):
                order of dimensions, accepts a combination of the characters
                'p' and 's', for processes and observations;
                must have the same length as number of dimensions in data
        """
        if len(dim_order) > 3:
            raise RuntimeError("dim_order can not have more than two " "entries")
        if len(dim_order) != data.ndim:
            raise RuntimeError(
                "Data array dimension ({0}) and length of "
                "dim_order ({1}) are not equal.".format(data.ndim, len(dim_order))
            )

        # Bring data into the order processes x observations in a pandas dataframe.
        data = self._reorder_data(data, dim_order)

        if n_processes is not None:
            data = data[:n_processes]
        if n_observations is not None:
            data = data[:, :n_observations]

        if self.detrend:
            print(Fore.GREEN + "[1/2] De-trending the dataset...")
            try:
                data = detrend(data, axis=1)
            except ValueError as err:
                print(f"Could not detrend data: {err}")
        else:
            print(Fore.RED + "[1/2] Skipping detrending of the dataset...")

        if self.normalise:
            print(Fore.GREEN + "[2/2] Normalising (z-scoring) the dataset...\n")
            data = zscore(data, axis=1, nan_policy="omit", ddof=1)
        else:
            print(Fore.RED + "[2/2] Skipping normalisation of the dataset...\n")

        nans = np.isnan(data)
        if nans.any():
            raise ValueError(
                f"Dataset {name} contains non-numerics (NaNs) in processes: {np.unique(np.where(nans)[0])}."
            )

        self._data = data
        self.data_type = type(data[0, 00, 0])

        self._reset_data_size()

        if name is not None:
            self._name = name

        if verbose:
            print(
                f'Dataset "{name}" now has properties: {self.n_processes} processes, {self.n_observations} observations, {self.n_replications} '
                "replications"
            )

    def add_process(self, proc, verbose=False):
        """Appends a univariate process to the dataset.

        Args:
            proc (ndarray):
                Univariate process to add, must be an array the same size as existing ones.
        """
        proc = np.squeeze(proc)
        if not isinstance(proc, np.ndarray) or proc.ndim != 1:
            raise TypeError("Process must be a 1D numpy array")

        if hasattr(self, "_data"):
            try:
                self._data = np.append(
                    self._data, np.reshape(proc, (1, self.n_observations, 1)), axis=0
                )
            except IndexError:
                raise IndexError()
        else:
            self.set_data(proc, dim_order="s", verbose=verbose)

        self._reset_data_size()

    def remove_process(self, procs):
        try:
            self._data = np.delete(self._data, procs, axis=0)
        except IndexError:
            print(
                f"Process {procs} is out of bounds of multivariate"
                f" time-series data with size {self.data.n_processes}"
            )

        self._reset_data_size()

    def _reorder_data(self, data, dim_order):
        """Reorder data dimensions to processes x observations x realisations."""
        # add singletons for missing dimensions
        missing_dims = "psr"
        for dim in dim_order:
            missing_dims = missing_dims.replace(dim, "")
        for dim in missing_dims:
            data = np.expand_dims(data, data.ndim)
            dim_order += dim

        # reorder array dims if necessary
        if dim_order[0] != "p":
            ind_p = dim_order.index("p")
            data = data.swapaxes(0, ind_p)
            dim_order = utils.swap_chars(dim_order, 0, ind_p)
        if dim_order[1] != "s":
            data = data.swapaxes(1, dim_order.index("s"))

        return data

    def _reset_data_size(self):
        """Set the data size."""
        self.n_processes = self._data.shape[0]
        self.n_observations = self._data.shape[1]
        self.n_replications = self._data.shape[2]


def load_dataset(name):
    basedir = os.path.join(os.path.dirname(__file__), "data")
    if name == "forex":
        filename = "forex.npy"
        dim_order = "sp"
    elif name == "cml":
        filename = "cml.npy"
        dim_order = "sp"
    else:
        raise NameError(f"Unknown dataset: {name}.")
    return Data(data=os.path.join(basedir, filename), dim_order=dim_order)



import warnings
import sklearn.covariance as cov
from scipy import stats, signal
import numpy as np

from pyspi.base import Undirected, Signed, parse_bivariate, parse_multivariate


class Estimators(Undirected, Signed):
    """Base class for (functional) connectivity-based statistics

    Information on covariance estimators at: https://scikit-learn.org/stable/modules/covariance.html
    """

    name = "Covariance"
    labels = ["basic", "unordered", "linear", "undirected"]

    def __init__(self, kind, estimator="EmpiricalCovariance", squared=False):
        paramstr = f"_{estimator}"
        if squared:
            paramstr = "-sq" + paramstr
            self.labels = Estimators.labels + ["unsigned"]
            self.issigned = lambda: False
        else:
            self.labels = Estimators.labels + ["signed"]
        self.identifier = self.identifier + paramstr
        self._squared = squared
        self._estimator = estimator
        self._kind = kind

    def _from_cache(self, data):
        try:
            mycov = data.covariance[self._estimator]
        except (AttributeError, KeyError):
            z = data.to_numpy(squeeze=True).T

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                mycov = getattr(cov, self._estimator)().fit(z)
            try:
                data.covariance[self._estimator] = mycov
            except AttributeError:
                data.covariance = {self._estimator: mycov}
        return mycov

    @parse_multivariate
    def multivariate(self, data):
        mycov = self._from_cache(data)
        matrix = getattr(mycov, self._kind + "_")
        np.fill_diagonal(matrix, np.nan)
        if self._squared:
            return np.square(matrix)
        else:
            return matrix


class Covariance(Estimators):

    name = "Covariance"
    identifier = "cov"

    def __init__(self, estimator="EmpiricalCovariance", squared=False):
        super().__init__(kind="covariance", squared=squared, estimator=estimator)


class Precision(Estimators):

    name = "Precision"
    identifier = "prec"

    def __init__(self, estimator="EmpiricalCovariance", squared=False):
        super().__init__(kind="precision", squared=squared, estimator=estimator)


class CrossCorrelation(Undirected, Signed):

    name = "Cross correlation"
    labels = ["basic", "linear", "undirected", "temporal"]

    def __init__(self, squared=False, statistic="max", sigonly=True):
        self.identifier = "xcorr"
        self._squared = squared
        self._statistic = statistic
        self._sigonly = sigonly

        if self._squared:
            self.issigned = lambda: False
            self.identifier = self.identifier + "-sq"
            self.labels = CrossCorrelation.labels + ["unsigned"]
        else:
            self.labels = CrossCorrelation.labels + ["signed"]
        self.identifier += f"_{statistic}_sig-{sigonly}"

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):

        T = data.n_observations
        try:
            r_ij = data.xcorr[(i, j)]
        except (KeyError, AttributeError):
            x, y = data.to_numpy()[[i, j]]

            r_ij = np.squeeze(signal.correlate(x, y, "full"))
            r_ij = r_ij / x.std() / y.std() / (T - 1)

            # Truncate to T/4
            r_ij = r_ij[T - T // 4 : T + T // 4]

            try:
                data.xcorr[(i, j)] = r_ij
            except AttributeError:
                data.xcorr = {(i, j): r_ij}
            data.xcorr[(j, i)] = data.xcorr[(i, j)]

        # Truncate at first statistically significant zero (i.e., |r| <= 1.96/sqrt(T))
        if self._sigonly:
            N = len(r_ij) // 2
            fzf = np.where(np.abs(r_ij[len(r_ij) // 2 :]) <= 1.96 / np.sqrt(N))[0][0]
            fzr = np.where(np.abs(r_ij[: len(r_ij) // 2]) <= 1.96 / np.sqrt(N))[0][-1]
            r_ij = r_ij[N - fzr : N + fzf]

        if self._statistic == "max":
            if self._squared:
                return np.max(r_ij**2)
            else:
                return np.max(r_ij)
        elif self._statistic == "mean":
            if self._squared:
                return np.mean(r_ij**2)
            else:
                return np.mean(r_ij)
        else:
            raise TypeError(f"Unknown statistic: {self._statistic}")


class SpearmanR(Undirected, Signed):

    name = "Spearman's correlation coefficient"
    identifier = "spearmanr"
    labels = ["basic", "unordered", "rank", "linear", "undirected"]

    def __init__(self, squared=False):
        self._squared = squared
        if squared:
            self.issigned = lambda: False
            self.identifier = self.identifier + "-sq"
            self.labels += ["unsigned"]
        else:
            self.labels += ["signed"]

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        x, y = data.to_numpy()[[i, j]]
        if self._squared:
            return stats.spearmanr(x, y).correlation ** 2
        else:
            return stats.spearmanr(x, y).correlation


class KendallTau(Undirected, Signed):

    name = "Kendall's tau"
    identifier = "kendalltau"
    labels = ["basic", "unordered", "rank", "linear", "undirected"]

    def __init__(self, squared=False):
        self._squared = squared
        if squared:
            self.issigned = lambda: False
            self.identifier = self.identifier + "-sq"
            self.labels += ["unsigned"]
        else:
            self.labels += ["signed"]

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        x, y = data.to_numpy()[[i, j]]
        if self._squared:
            return stats.kendalltau(x, y).correlation ** 2
        else:
            return stats.kendalltau(x, y).correlation



import numpy as np
from scipy.stats import zscore
import warnings
import pandas as pd
import os
import yaml 
from colorama import Fore, init
init(autoreset=True)

def _contains_nan(a, nan_policy='propagate'):
    policies = ['propagate', 'raise', 'omit']
    if nan_policy not in policies:
        raise ValueError("nan_policy must be one of {%s}" %
                         ', '.join("'%s'" % s for s in policies))
    try:
        # Calling np.sum to avoid creating a huge array into memory
        # e.g. np.isnan(a).any()
        with np.errstate(invalid='ignore'):
            contains_nan = np.isnan(np.sum(a))
    except TypeError:
        # If the check cannot be properly performed we fallback to omitting
        # nan values and raising a warning. This can happen when attempting to
        # sum things that are not numbers (e.g. as in the function `mode`).
        contains_nan = False
        nan_policy = 'omit'
        warnings.warn("The input array could not be properly checked for nan "
                      "values. nan values will be ignored.", RuntimeWarning)

    if contains_nan and nan_policy == 'raise':
        raise ValueError("The input contains nan values")

    return (contains_nan, nan_policy)


def strshort(instr,mlength):
    """Shorten a string using ellipsis
    """
    if isinstance(instr,list):
        outstr = []
        for i in range(len(instr)):
            cstr = instr[i]
            outstr.append((cstr[:mlength-6] + '...' + cstr[-3:]) if len(cstr) > mlength else cstr)
    else:
        outstr = (instr[:mlength-6] + '...' + instr[-3:]) if len(instr) > mlength else instr
    return outstr

def acf(x,mode='positive'):
    """Return the autocorrelation function
    """
    if x.ndim > 1:
        x = np.squeeze(x)

    x = zscore(x)
    acf = np.correlate(x,x,mode='full')
    acf = acf / acf[acf.size//2]

    if mode == 'positive':
        return acf[acf.size//2:]
    else:
        return acf

def swap_chars(s, i_1, i_2):
    """Swap to characters in a string.

    Example:
        >>> print(swap_chars('heLlotHere', 2, 6))
        'heHlotLere'
    """
    if i_1 > i_2:
        i_1, i_2 = i_2, i_1
    return ''.join([s[0:i_1], s[i_2], s[i_1+1:i_2], s[i_1], s[i_2+1:]])

def normalise(a, axis=0, nan_policy='propogate'):

    contains_nan, nan_policy = _contains_nan(a, nan_policy)

    if contains_nan and nan_policy == 'omit':
        return (a - np.nanmin(a,axis=axis)) / (np.nanmax(a,axis=axis) - np.nanmin(a,axis=axis))
    else:
        return (a - np.min(a,axis=axis)) / (np.max(a,axis=axis) - np.min(a,axis=axis))

def standardise(a, dimension=0, df=1):
    """Z-standardise a numpy array along a given dimension.

    Standardise array along the axis defined in dimension using the denominator
    (N - df) for the calculation of the standard deviation.

    Args:
        a : numpy array
            data to be standardised
        dimension : int [optional]
            dimension along which array should be standardised
        df : int [optional]
            degrees of freedom for the denominator of the standard derivation

    Returns:
        numpy array
            standardised data
    """
    # Avoid division by standard deviation if the process is constant.
    a_sd = a.std(axis=dimension, ddof=df)

    if np.isclose(a_sd, 0):
        return a - a.mean(axis=dimension)
    else:
        return (a - a.mean(axis=dimension)) / a_sd

def convert_mdf_to_ddf(df):
    ddf = pd.pivot_table(data=df.stack(dropna=False).reset_index(),index='Dataset',columns=['SPI-1', 'SPI-2'],dropna=False).T.droplevel(0)
    return ddf

def is_jpype_jvm_available():
    """Check whether a JVM is accessible via Jpype"""
    try:
        import jpype as jp
        if not jp.isJVMStarted():
            jarloc = (os.path.dirname(os.path.abspath(__file__)) + "/lib/jidt/infodynamics.jar")
            # if JVM not started, start a session
            print(f"Starting JVM with java class {jarloc}.")
            jp.startJVM(jp.getDefaultJVMPath(), "-ea", "-Djava.class.path=" + jarloc)
        return True
    except Exception as e:
        print(f"Jpype JVM not available: {e}")
        return False

def is_octave_available():
    """Check whether octave is available"""
    try:
        from oct2py import Oct2Py
        oc = Oct2Py()
        oc.exit()
        return True
    except Exception as e:
        print(f"Octave not available: {e}")
        return False

def check_optional_deps():
    """Bundle all of the optional
    dependency checks together."""
    isAvailable = {}
    isAvailable['octave'] = is_octave_available()
    isAvailable['java'] = is_jpype_jvm_available()

    return isAvailable

def filter_spis(keywords, output_name = None, configfile= None):
    """
    Filter a YAML using a list of keywords, and save the reduced set as a new
    YAML with a user-specified name (or a random one if not provided) in the
    current directory.

    Args:
        keywords (list): A list of keywords (as strings) to filter the YAML.
        output_name (str, optional): The desired name for the output file. Defaults to a random name. 
        configfile (str, optional): The path to the input YAML file. Defaults to the `config.yaml' in the pyspi dir. 

    Raises:
        ValueError: If `keywords` is not a list or if no SPIs match the keywords.
        FileNotFoundError: If the specified `configfile` or the default `config.yaml` is not found.
        IOError: If there's an error reading the YAML file.
    """
    # handle invalid keyword input
    if not keywords:
        raise ValueError("At least one keyword must be provided.")
    if not all(isinstance(keyword, str) for keyword in keywords):
        raise ValueError("All keywords must be strings.")
    if not isinstance(keywords, list):
        raise ValueError("Keywords must be provided as a list of strings.")

    # if no configfile and no keywords are provided, use the default 'config.yaml' in pyspi location
    if configfile is None:
        script_dir = os.path.dirname(os.path.abspath(__file__))
        default_config = os.path.join(script_dir, 'config.yaml')
        if not os.path.isfile(default_config):
            raise FileNotFoundError(f"Default 'config.yaml' file not found in {script_dir}.")
        configfile = default_config
        source_file_info = f"Default 'config.yaml' file from {script_dir} was used as the source file."
    else:
        source_file_info = f"User-specified config file '{configfile}' was used as the source file."

    # load in user-specified yaml
    try:
        with open(configfile) as f:
            yf = yaml.load(f, Loader=yaml.FullLoader)
    except FileNotFoundError:
        raise FileNotFoundError(f"Config file '{configfile}' not found.")
    except Exception as e:
        # handle all other exceptions
        raise IOError(f"An error occurred while trying to read '{configfile}': {e}")

    # new dictionary to be converted to final YAML
    filtered_subset = {}
    spis_found = 0
    
    for module in yf:
        module_spis = {}
        for spi in yf[module]:
            spi_labels = yf[module][spi].get('labels')
            if all(keyword in spi_labels for keyword in keywords):
                module_spis[spi] = yf[module][spi]
                if yf[module][spi].get('configs'):
                    spis_found += len(yf[module][spi].get('configs'))
                else:
                    spis_found += 1
    
        if module_spis:
            filtered_subset[module] = module_spis
    
    # check that > 0 SPIs found
    if spis_found == 0:
        raise ValueError(f"0 SPIs were found with the specific keywords: {keywords}.")
    
    # construct output file path
    if output_name is None:
        # use a unique name
        output_name = "config_" + os.urandom(4).hex()

    output_file = os.path.join(os.getcwd(), f"{output_name}.yaml")
    
    # write to YAML
    with open(output_file, "w") as outfile:
        yaml.dump(filtered_subset, outfile, default_flow_style=False, sort_keys=False)

    # output relevant information
    print(f"""\nOperation Summary:
-----------------
- {source_file_info}
- Total SPIs Matched: {spis_found} SPI(s) were found with the specific keywords: {keywords}.
- New File Created: A YAML file named `{output_name}.yaml` has been saved in the current directory: `{output_file}'
- Next Steps: To utilise the filtered set of SPIs, please initialise a new Calculator instance with the following command:
`Calculator(configfile='{output_file}')`
""")

def inspect_calc_results(calc):
    """
    Display a summary of the computed SPI results, including counts of successful computations, 
    outputs with NaNs, and partially computed results.
    """
    total_num_spis = calc.n_spis
    num_procs = calc.dataset.n_processes
    spi_results = dict({'Successful': list(), 'NaNs': list(), 'Partial NaNs': list()})
    for key in calc.spis.keys():
        if calc.table[key].isna().all().all():
            spi_results['NaNs'].append(key)
        elif calc.table[key].isnull().values.sum() > num_procs:
            # off-diagonal NaNs
            spi_results['Partial NaNs'].append(key)
        else:
            # returned numeric values (i.e., not NaN)
            spi_results['Successful'].append(key)
    
    # print summary
    double_line_60 = "="*60
    single_line_60 = "-"*60
    print("\nSPI Computation Results Summary")
    print(double_line_60)
    print(f"\nTotal number of SPIs attempted: {total_num_spis}")
    print(f"Number of SPIs successfully computed: {len(spi_results['Successful'])} ({len(spi_results['Successful']) / total_num_spis * 100:.2f}%)")
    print(single_line_60)
    print("Category       | Count | Percentage")
    print(single_line_60)
    for category, spis in spi_results.items():
        count = len(spis)
        percentage = (count / total_num_spis) * 100
        print(f"{category:14} | {count:5} | {percentage:6.2f}%")
    print(single_line_60)

    if spi_results['NaNs']:
        print(f"\n[{len(spi_results['NaNs'])}] SPI(s) produced NaN outputs:")
        print(single_line_60)
        for i, spi in enumerate(spi_results['NaNs']):
            print(f"{i+1}. {spi}")
        print(single_line_60 + "\n")
    if spi_results['Partial NaNs']:
        print(f"\n[{len(spi_results['Partial NaNs'])}] SPIs which produced partial NaN outputs:")
        print(single_line_60)
        for i, spi in enumerate(spi_results['Partial NaNs']):
            print(f"{i+1}. {spi}")
        print(single_line_60 + "\n")
    


import numpy as np
from pyspi.data import Data
import warnings, copy

"""
Some parsing functions for decorating so that we can either input the time series directly or use the data structure
"""
def parse_univariate(function):
    def parsed_function(self,data,i=None,inplace=True):
        if not isinstance(data,Data):
            data1 = data
            data = Data(data=data1)
        elif not inplace:
            # Ensure we don't write over the original
            data = copy.deepcopy(data)

        if i is None:
            if data.n_processes == 1:
                i = 0
            else:
                raise ValueError('Require argument i to be set.')

        return function(self,data,i=i)

    return parsed_function

def parse_bivariate(function):
    def parsed_function(self,data,data2=None,i=None,j=None,inplace=True):
        if not isinstance(data,Data):
            if data2 is None:
                raise TypeError('Input must be either a pyspi.data object or two 1D-array inputs.'
                                    f' Received {type(data)} and {type(data2)}.')
            data1 = data
            data = Data()
            data.add_process(data1)
            data.add_process(data2)
        elif not inplace:
            # Ensure we don't write over the original
            data = copy.deepcopy(data)

        if i is None and j is None:
            if data.n_processes == 2:
                i,j = 0,1
            else:
                Warning('i and j not set.')

        return function(self,data,i=i,j=j)

    return parsed_function

def parse_multivariate(function):
    def parsed_function(self,data,inplace=True):
        if not isinstance(data,Data):
            # Create a pyspi.Data object from iterable data object
            try:
                procs = data
                data = Data()
                for p in procs:
                    data.add_process(p)
            except IndexError:
                raise TypeError('Data must be either a pyspi.data.Data object or an and iterable of numpy.ndarray''s.')
        elif not inplace:
            # Ensure we don't write over the original
            data = copy.deepcopy(data)

        return function(self,data)

    return parsed_function

class Directed:
    """ Base class for directed statistics
    """

    name = 'Bivariate base class'
    identifier = 'bivariate_base'
    labels = ['signed']

    @parse_bivariate
    def bivariate(self,data,i=None,j=None):
        """ Overload method for getting the pairwise dependencies
        """
        raise NotImplementedError("Method not yet overloaded.")

    @parse_multivariate
    def multivariate(self,data):
        """ Compute the dependency statistics for the entire multivariate dataset
        """
        A = np.empty((data.n_processes,data.n_processes))
        A[:] = np.nan

        for j in range(data.n_processes):
            for i in [ii for ii in range(data.n_processes) if ii != j]:
                A[i,j] = self.bivariate(data,i=i,j=j)
        return A

    def get_group(self,classes):
        for i, i_cls in enumerate(classes):
            for j, j_cls in enumerate(classes):
                if i == j:
                    continue
                assert not set(i_cls).issubset(set(j_cls)), (f'Class {i_cls} is a subset of class {j_cls}.')

        self._group = None
        self._group_name = None

        labset = set(self.labels)
        matches = [set(cls).issubset(labset) for cls in classes]

        if np.count_nonzero(matches) > 1:
            warnings.warn(f'More than one match for classes {classes}')
        else:
            try:
                id = np.where(matches)[0][0]
                self._group = id
                self._group_name = ', '.join(classes[id])
                return self._group, self._group_name
            except (TypeError,IndexError):
                pass
        return None

class Undirected(Directed):
    """ Base class for directed statistics
    """

    name = 'Base class'
    identifier = 'base'
    labels = ['unsigned']

    def ispositive(self):
        return False

    @parse_multivariate
    def multivariate(self,data):
        A = super(Undirected,self).multivariate(data)

        li = np.tril_indices(data.n_processes,-1)
        A[li] = A.T[li]
        return A

class Signed:
    """ Base class for signed SPIs
    """
    def issigned(self):
        return True

class Unsigned:
    """ Base class for unsigned SPIs
    """
    def issigned(self):
        return False


import numpy as np
from copy import deepcopy

import spectral_connectivity as sc  # For directed spectral statistics (excl. spectral GC)
from pyspi.base import (
    Directed,
    Undirected,
    Unsigned,
    parse_bivariate,
    parse_multivariate,
)
import nitime.analysis as nta
import nitime.timeseries as ts
import warnings


class NonparametricSpectral(Unsigned):
    """Base class for the nonparametric spectral methods from the Eden-Kramer repo"""

    def __init__(self, fs=1, fmin=0, fmax=None, statistic="mean"):
        if fmax is None:
            fmax = fs / 2

        self._fs = fs
        if fs != 1:
            warnings.warn("Multiple sampling frequencies not yet handled.")
        self._fmin = fmin
        self._fmax = fmax
        if statistic == "mean":
            self._statfn = np.nanmean
        elif statistic == "max":
            self._statfn = np.nanmax
        elif statistic not in ["delay", "slope", "rvalue"]:
            raise NameError(f"Unknown statistic: {statistic}")
        else:
            self._statfn = None
        self._statistic = statistic
        paramstr = (
            f"_multitaper_{statistic}_fs-{fs}_fmin-{fmin:.3g}_fmax-{fmax:.3g}".replace(
                ".", "-"
            )
        )
        self.identifier += paramstr

    @property
    def key(self):
        if isinstance(self, GroupDelay) or isinstance(self, PhaseSlopeIndex):
            return (self.measure, self._fmin, self._fmax)
        else:
            return (self.measure,)

    @property
    def measure(self):
        try:
            return self._measure
        except AttributeError:
            raise AttributeError(f"Include measure for {self.identifier}")

    def _get_statistic(self, C):
        raise NotImplementedError


class NonparametricSpectralMultivariate(NonparametricSpectral):
    def _get_cache(self, data):
        try:
            res = data.spectral_mv[self.key]
            freq = data.spectral_mv["freq"]
        except (AttributeError, KeyError):
            z = np.transpose(data.to_numpy(squeeze=True))
            m = sc.Multitaper(z, sampling_frequency=self._fs)
            conn = sc.Connectivity.from_multitaper(m)
            try:
                res = getattr(conn, self.measure)()
            except TypeError:
                res = self._get_statistic(conn)

            freq = conn.frequencies
            try:
                data.spectral_mv[self.key] = res
            except AttributeError:
                data.spectral_mv = {"freq": freq, self.measure: res}

        return res, freq

    @parse_multivariate
    def multivariate(self, data):
        adj_freq, freq = self._get_cache(data)
        freq_id = np.where((freq >= self._fmin) * (freq <= self._fmax))[0]

        try:
            adj = self._statfn(adj_freq[0, freq_id, :, :], axis=0)
        except IndexError:  # For phase slope index
            adj = adj_freq[0]
        except TypeError:  # For group delay
            stat_id = [
                i
                for i, s in enumerate(["delay", "slope", "rvalue"])
                if self._statistic == s
            ][0]
            adj = adj_freq[stat_id][0]
        np.fill_diagonal(adj, np.nan)
        return adj


class NonparametricSpectralBivariate(NonparametricSpectral):
    def _get_cache(self, data, i, j):
        key = (self.measure, i, j)
        try:
            res = data.spectral_bv[key]
            freq = data.spectral_bv["freq"]
        except (KeyError, AttributeError):
            z = np.transpose(data.to_numpy(squeeze=True)[[i, j]])
            m = sc.Multitaper(z, sampling_frequency=self._fs)
            conn = sc.Connectivity.from_multitaper(m)
            try:
                res = getattr(conn, self.measure)()
            except TypeError:
                res = self._get_statistic(conn)

            freq = conn.frequencies
            try:
                data.spectral_bv[key] = res
            except AttributeError:
                data.spectral_bv = {"freq": freq, key: res}

        return res, freq

    @parse_bivariate
    def bivariate(self, data, i=None, j=None):
        """TODO: cache this result"""
        bv_freq, freq = self._get_cache(data, i, j)
        freq_id = np.where((freq > self._fmin) * (freq < self._fmax))[0]

        return self._statfn(bv_freq[0, freq_id, 0, 1])


class CoherenceMagnitude(NonparametricSpectralMultivariate, Undirected):
    name = "Coherence magnitude"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "cohmag"
        super().__init__(**kwargs)
        self._measure = "coherence_magnitude"


class CoherencePhase(NonparametricSpectralMultivariate, Undirected):
    name = "Coherence phase"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "phase"
        super().__init__(**kwargs)
        self._measure = "coherence_phase"


class ImaginaryCoherence(NonparametricSpectralMultivariate, Undirected):
    name = "Imaginary coherence"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "icoh"
        super().__init__(**kwargs)
        self._measure = "imaginary_coherence"


class PhaseLockingValue(NonparametricSpectralMultivariate, Undirected):
    name = "Phase locking value"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "plv"
        super().__init__(**kwargs)
        self._measure = "phase_locking_value"

        myfn = deepcopy(self._statfn)
        self._statfn = lambda x, **kwargs: myfn(np.absolute(x), **kwargs)


class PhaseLagIndex(NonparametricSpectralMultivariate, Undirected):
    name = "Phase lag index"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "pli"
        super().__init__(**kwargs)
        self._measure = "phase_lag_index"


class WeightedPhaseLagIndex(NonparametricSpectralMultivariate, Undirected):
    name = "Weighted phase lag index"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "wpli"
        super().__init__(**kwargs)
        self._measure = "weighted_phase_lag_index"


class DebiasedSquaredPhaseLagIndex(NonparametricSpectralMultivariate, Undirected):
    name = "Debiased squared phase lag index"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "dspli"
        super().__init__(**kwargs)
        self._measure = "debiased_squared_phase_lag_index"


class DebiasedSquaredWeightedPhaseLagIndex(
    NonparametricSpectralMultivariate, Undirected
):
    name = "Debiased squared weighted phase-lag index"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "dswpli"
        super().__init__(**kwargs)
        self._measure = "debiased_squared_weighted_phase_lag_index"


class PairwisePhaseConsistency(NonparametricSpectralMultivariate, Undirected):
    name = "Pairwise phase consistency"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "ppc"
        super().__init__(**kwargs)
        self._measure = "pairwise_phase_consistency"


"""
    These next several seem to segfault for large vector autoregressive processes (something to do with np.linalg solver).
    Switched them to bivariate for now until the issue is resolved
"""


class DirectedCoherence(NonparametricSpectralBivariate, Directed):
    name = "Directed coherence"
    labels = ["unsigned", "spectral", "directed"]

    def __init__(self, **kwargs):
        self.identifier = "dcoh"
        super().__init__(**kwargs)
        self._measure = "directed_coherence"


class PartialDirectedCoherence(NonparametricSpectralBivariate, Directed):
    name = "Partial directed coherence"
    labels = ["unsigned", "spectral", "directed"]

    def __init__(self, **kwargs):
        self.identifier = "pdcoh"
        super().__init__(**kwargs)
        self._measure = "partial_directed_coherence"


class GeneralizedPartialDirectedCoherence(NonparametricSpectralBivariate, Directed):
    name = "Generalized partial directed coherence"
    labels = ["unsigned", "spectral", "directed"]

    def __init__(self, **kwargs):
        self.identifier = "gpdcoh"
        super().__init__(**kwargs)
        self._measure = "generalized_partial_directed_coherence"


class DirectedTransferFunction(NonparametricSpectralBivariate, Directed):
    name = "Directed transfer function"
    labels = ["unsigned", "spectral", "directed", "lagged"]

    def __init__(self, **kwargs):
        self.identifier = "dtf"
        super().__init__(**kwargs)
        self._measure = "directed_transfer_function"


class DirectDirectedTransferFunction(NonparametricSpectralBivariate, Directed):
    name = "Direct directed transfer function"
    labels = ["unsigned", "spectral", "directed", "lagged"]

    def __init__(self, **kwargs):
        self.identifier = "ddtf"
        super().__init__(**kwargs)
        self._measure = "direct_directed_transfer_function"


class PhaseSlopeIndex(NonparametricSpectralMultivariate, Undirected):
    name = "Phase slope index"
    labels = ["unsigned", "spectral", "undirected"]

    def __init__(self, **kwargs):
        self.identifier = "psi"
        super().__init__(**kwargs)
        self._measure = "phase_slope_index"

    def _get_statistic(self, C):
        return C.phase_slope_index(
            frequencies_of_interest=[self._fmin, self._fmax],
            frequency_resolution=(self._fmax - self._fmin) / 50,
        )


class GroupDelay(NonparametricSpectralMultivariate, Directed):
    name = "Group delay"
    labels = ["unsigned", "spectral", "directed", "lagged"]

    def __init__(self, **kwargs):
        self.identifier = "gd"
        super().__init__(**kwargs)
        self._measure = "group_delay"

    def _get_statistic(self, C):
        return C.group_delay(
            frequencies_of_interest=[self._fmin, self._fmax],
            frequency_resolution=(self._fmax - self._fmin) / 50,
        )


class SpectralGrangerCausality(NonparametricSpectralMultivariate, Directed, Unsigned):
    name = "Spectral Granger causality"
    identifier = "sgc"
    labels = ["unsigned", "embedding", "spectral", "directed", "lagged"]

    def __init__(
        self,
        fs=1,
        fmin=1e-5,
        fmax=0.5,
        method="nonparametric",
        order=None,
        max_order=50,
        statistic="mean",
        ignore_nan=True,
        nan_threshold=0.5,
    ):
        self._fs = fs  # Not yet implemented
        self._fmin = fmin
        self._fmax = fmax
        self.ignore_nan = ignore_nan
        self.nan_threshold = nan_threshold

        if self._fmin <= 0.0:
            warnings.warn(f"Frequency minimum set to {self._fmin}; overriding to 1e-5.")
            self._fmin = 1e-5

        if statistic == "mean":
            if self.ignore_nan:
                self._statfn = np.nanmean
            else:
                self._statfn = np.mean
        elif statistic == "max":
            if self.ignore_nan:
                self._statfn = np.nanmax
            else:
                self._statfn = np.max
        else:
            raise NameError(f"Unknown statistic {statistic}")

        self._method = method
        if self._method == "nonparametric":
            self._measure = "pairwise_spectral_granger_prediction"
            paramstr = f"_nonparametric_{statistic}_fs-{fs}_fmin-{fmin:.3g}_fmax-{fmax:.3g}".replace(
                ".", "-"
            )
        else:
            self._order = order
            self._max_order = max_order
            paramstr = f"_parametric_{statistic}_fs-{fs}_fmin-{fmin:.3g}_fmax-{fmax:.3g}_order-{order}".replace(
                ".", "-"
            )

        self.identifier = self.identifier + paramstr

    def _getkey(self):
        if self._method == "nonparametric":
            return (self._method, -1, -1)
        else:
            return (self._method, self._order, self._max_order)

    def _get_cache(self, data):
        key = self._getkey()

        try:
            F = data.spectral_gc[key]["F"]
            freq = data.spectral_gc[key]["freq"]
        except (AttributeError, KeyError):

            if self._method == "nonparametric":
                F, freq = super()._get_cache(data)
            else:
                z = data.to_numpy(squeeze=True)
                time_series = ts.TimeSeries(z, sampling_interval=1)
                GA = nta.GrangerAnalyzer(
                    time_series, order=self._order, max_order=self._max_order
                )

                triu_id = np.triu_indices(data.n_processes)

                F = np.full(GA.causality_xy.shape, np.nan)
                F[triu_id[0], triu_id[1], :] = GA.causality_xy[
                    triu_id[0], triu_id[1], :
                ]
                F[triu_id[1], triu_id[0], :] = GA.causality_yx[
                    triu_id[0], triu_id[1], :
                ]

                F = np.transpose(np.expand_dims(F, axis=3), axes=[3, 2, 1, 0])
                freq = GA.frequencies
            try:
                data.spectral_gc[key] = {"freq": freq, "F": F}
            except AttributeError:
                data.spectral_gc = {key: {"freq": freq, "F": F}}

        return F, freq

    @parse_multivariate
    def multivariate(self, data):
        try:
            cache, freq = self._get_cache(data)
            freq_id = np.where((freq >= self._fmin) * (freq <= self._fmax))[0]

            result = self._statfn(cache[0, freq_id, :, :], axis=0)

            nan_pct = np.isnan(cache[0, freq_id, :, :]).mean(axis=0)
            np.fill_diagonal(nan_pct, 0.0)

            isna = nan_pct > self.nan_threshold
            if isna.any():
                warnings.warn(
                    f"Spectral GC: the following processes have >{self.nan_threshold*100:.1f}% "
                    f"NaN values:\n{np.transpose(np.where(isna))}\nThese indices will be set to NaN. "
                    "Set ignore_nan to False or modify nan_threshold parameter if required."
                )
                result[isna] = np.nan

            return result
        except ValueError as err:
            warnings.warn(err)
            return np.full((data.n_processes, data.n_processes), np.nan)



# Science/maths/computing tools
import numpy as np
import pandas as pd
import copy, yaml, importlib, time, warnings, os
from tqdm import tqdm
from collections import Counter
from scipy import stats
from colorama import init, Fore
init(autoreset=True)

# From this package
from .data import Data
from .utils import convert_mdf_to_ddf, check_optional_deps, inspect_calc_results


class Calculator:
    """Compute all pairwise interactions.

    The calculator takes in a multivariate time-series dataset, computes and stores all pairwise interactions for the dataset.
    It uses a YAML configuration file that can be modified in order to compute a reduced set of pairwise methods.

    Example:
        >>> import numpy as np
        >>> dataset = np.random.randn(5,500)    # create a random multivariate time series (MTS)
        >>> calc = Calculator(dataset=dataset)  # Instantiate the calculator
        >>> calc.compute()                      # Compute all pairwise interactions

    Args:
        dataset (:class:`~pyspi.data.Data`, array_like, optional):
            The multivariate time series of M processes and T observations, defaults to None.
        name (str, optional):
            The name of the calculator. Mainly used for printing the results but can be useful if you have multiple instances, defaults to None.
        labels (array_like, optional):
            Any set of strings by which you want to label the calculator. This can be useful later for classification purposes, defaults to None.
        subset (str, optional):
            A pre-configured subset of SPIs to use. Options are "all", "fast", "sonnet", or "fabfour", defaults to "all".
        configfile (str, optional):
            The location of the YAML configuration file for a user-defined subset. See :ref:`Using a reduced SPI set`, defaults to :code:`'</path/to/pyspi>/pyspi/config.yaml'`
        detrend (bool, optional):
            If True, detrend the dataset along the time axis before normalising (if enabled), defaults to True.
        normalise (bool, optional):
            If True, z-score normalise the dataset along the time axis before computing SPIs, defaults to True.
            Detrending (if enabled) is always applied before normalisation.
    """
    _optional_dependencies = None

    def __init__(
        self, dataset=None, name=None, labels=None, subset="all", configfile=None,
        detrend=True, normalise=True
    ):
        self._spis = {}
        self._excluded_spis = list()
        self._normalise = normalise
        self._detrend = detrend

        # Define configfile by subset if it was not specified
        if configfile is None:
            if subset == "fast":
                configfile = (
                    os.path.dirname(os.path.abspath(__file__)) + "/fast_config.yaml"
                )
            elif subset == "sonnet":
                configfile = (
                    os.path.dirname(os.path.abspath(__file__)) + "/sonnet_config.yaml"
                )
            elif subset == "fabfour":
                configfile = (
                    os.path.dirname(os.path.abspath(__file__)) + "/fabfour_config.yaml"
                )
            # If no configfile was provided but the subset was not one of the above (or the default 'all'), raise an error
            elif subset != "all":
                raise ValueError(
                    f"Subset '{subset}' does not exist. Try 'all' (default), 'fast', 'sonnet', or 'fabfour'."
                )
            else:
                configfile = os.path.dirname(os.path.abspath(__file__)) + "/config.yaml"

        # add dependency checks here if the calculator is being instantiated for the first time
        if not Calculator._optional_dependencies:
            # check if optional dependencies exist
            print("Checking if optional dependencies exist...")
            Calculator._optional_dependencies = check_optional_deps()

        self._load_yaml(configfile)

        duplicates = [
            name for name, count in Counter(self._spis.keys()).items() if count > 1
        ]
        if len(duplicates) > 0:
            raise ValueError(
                f"Duplicate SPI identifiers: {duplicates}.\n Check the config file for duplicates."
            )

        self._name = name
        self._labels = labels

        print(f"="*100)
        print(Fore.GREEN + f"{len(self.spis)} SPI(s) were successfully initialised.\n")
        if len(self._excluded_spis) > 0:
            missing_deps = [dep for dep, is_met in self._optional_dependencies.items() if not is_met]
            print(Fore.YELLOW + "**** SPI Initialisation Warning ****")
            print(Fore.YELLOW + "\nSome dependencies were not detected, which has led to the exclusion of certain SPIs:")
            print("\nMissing Dependencies:")

            for dep in missing_deps:
                print(f"- {dep}")

            print(f"\nAs a result, a total of {len(self._excluded_spis)} SPI(s) have been excluded:\n")

            dependency_groups = {}
            for spi in self._excluded_spis:
                for dep in spi[1]: 
                    if dep not in dependency_groups:
                        dependency_groups[dep] = []
                    dependency_groups[dep].append(spi[0])

            for dep, spis in dependency_groups.items():
                print(f"\nDependency - {dep} - affects {len(spis)} SPI(s)")
                print("Excluded SPIs:")
                for spi in spis:
                    print(f"  - {spi}")

            print(f"\n" + "="*100)
            print(Fore.YELLOW + "\nOPTIONS TO PROCEED:\n")
            print(f"  1) Install the following dependencies to access all SPIs: [{', '.join(missing_deps)}]")
            callable_name = "{Calculator/CalculatorFrame}"
            print(f"  2) Continue with a reduced set of {self.n_spis} SPIs by calling {callable_name}.compute(). \n")
            print(f"="*100 + "\n")

        if dataset is not None:
            self.load_dataset(dataset)

    @property
    def spis(self):
        """Dict of SPIs.

        Keys are the SPI identifier and values are their objects.
        """
        return self._spis

    @spis.setter
    def spis(self, s):
        raise Exception("Do not set this property externally.")

    @property
    def n_spis(self):
        """Number of SPIs in the calculator."""
        return len(self._spis)

    @property
    def dataset(self):
        """Dataset as a data object."""
        return self._dataset

    @dataset.setter
    def dataset(self, d):
        raise Exception(
            "Do not set this property externally. Use the load_dataset() method."
        )

    @property
    def name(self):
        """Name of the calculator."""
        return self._name

    @name.setter
    def name(self, n):
        self._name = n

    @property
    def labels(self):
        """List of calculator labels."""
        return self._labels

    @labels.setter
    def labels(self, ls):
        self._labels = ls

    @property
    def table(self):
        """Results table for all pairwise interactions."""
        return self._table

    @table.setter
    def table(self, a):
        raise Exception(
            "Do not set this property externally. Use the compute() method."
        )

    @property
    def group(self):
        """The numerical group assigned during :meth:`~pyspi.Calculator.calculator.set_group`."""
        try:
            return self._group
        except AttributeError as err:
            warnings.warn("Group undefined. Call set_group() method first.")
            raise AttributeError(err)

    @group.setter
    def group(self, g):
        raise Exception(
            "Do not set this property externally. Use the set_group() method."
        )

    @property
    def group_name(self):
        """The group name assigned during :meth:`~pyspi.Calculator.calculator.set_group`."""
        try:
            return self._group_name
        except AttributeError as err:
            warnings.warn(f"Group name undefined. Call set_group() method first.")
            return None

    @group_name.setter
    def group_name(self, g):
        raise Exception("Do not set this property externally. Use the group() method.")

    def _load_yaml(self, document):
        print("Loading configuration file: {}".format(document))

        with open(document) as f:
            yf = yaml.load(f, Loader=yaml.FullLoader)

            # Instantiate the SPIs
            for module_name in yf:
                print("*** Importing module {}".format(module_name))
                module = importlib.import_module(module_name, __package__)
                for fcn in yf[module_name]:
                    deps = yf[module_name][fcn].get('dependencies')
                    if deps is not None:
                        all_deps_met = all(Calculator._optional_dependencies.get(dep, False) for dep in deps)
                        if not all_deps_met:
                            current_base_spi = yf[module_name][fcn]
                            print(f"Optional dependencies: {deps} not met. Skipping {len(current_base_spi.get('configs'))} SPI(s):")
                            for params in current_base_spi.get('configs'):
                                print(f"*SKIPPING SPI: {module_name}.{fcn}(x,y,{params})...")
                                self._excluded_spis.append([f"{fcn}(x,y,{params})", deps])
                            continue
                    try:
                        for params in yf[module_name][fcn].get('configs'):
                            print(
                                f"[{self.n_spis}] Adding SPI {module_name}.{fcn}(x,y,{params})"
                            )
                            spi = getattr(module, fcn)(**params)
                            self._spis[spi.identifier] = spi
                            print(
                                f'Succesfully initialised SPI with identifier "{spi.identifier}" and labels {spi.labels}'
                            )
                    except TypeError:
                        print(f"[{self.n_spis}] Adding SPI {module_name}.{fcn}(x,y)...")
                        spi = getattr(module, fcn)()
                        self._spis[spi.identifier] = spi
                        print(
                            f'Succesfully initialised SPI with identifier "{spi.identifier}" and labels {spi.labels}'
                        )

    def load_dataset(self, dataset):
        """Load new dataset into existing instance.

        Args:
            dataset (:class:`~pyspi.data.Data`, array_list):
                New dataset to attach to calculator.
        """
        if not isinstance(dataset, Data):
            self._dataset = Data(Data.convert_to_numpy(dataset), normalise=self._normalise, detrend=self._detrend)
        else:
            self._dataset = dataset

        columns = pd.MultiIndex.from_product(
            [self.spis.keys(), self._dataset.procnames], names=["spi", "process"]
        )
        self._table = pd.DataFrame(
            data=np.full(
                (self.dataset.n_processes, self.n_spis * self.dataset.n_processes),
                np.nan,
            ),
            columns=columns,
            index=self._dataset.procnames,
        )
        self._table.columns.name = "process"

    def compute(self):
        """Compute the SPIs on the MVTS dataset."""
        if not hasattr(self, "_dataset"):
            raise AttributeError(
                "Dataset not loaded yet. Please initialise with load_dataset."
            )

        pbar = tqdm(self.spis.keys())
        for spi in pbar:
            pbar.set_description(f"Processing [{self._name}: {spi}]")
            start_time = time.time()
            try:
                # Get the MPI from the dataset
                S = self._spis[spi].multivariate(self.dataset)

                # Ensure the diagonal is NaN (sometimes set within the functions)
                np.fill_diagonal(S, np.nan)

                # Save results
                self._table[spi] = S
            except Exception as err:
                warnings.warn(f'Caught {type(err)} for SPI "{spi}": {err}')
                self._table[spi] = np.nan
        pbar.close()
        print(Fore.GREEN + f"\nCalculation complete. Time taken: {pbar.format_dict['elapsed']:.4f}s")
        inspect_calc_results(self)
        
    def _rmmin(self):
        """Iterate through all spis and remove the minimum (fixes absolute value errors when correlating)"""
        for spi in self.spis:
            mpi = self.table[spi]
            if not self.spis[spi].issigned():
                self.table[spi] = mpi - np.nanmin(mpi)

    def set_group(self, classes):
        """Assigns a numeric value to this instance based on list of classes.

        Args:
            classes (list):
                If any of the labels in this instance matches one in the class list, then we assign the index
                value to this class.
        """
        self._group = None
        self._group_name = None

        # Ensure this is a list of lists
        for i, c in enumerate(classes):
            if not isinstance(c, list):
                classes[i] = [c]

        for i, i_cls in enumerate(classes):
            for j, j_cls in enumerate(classes):
                if i == j:
                    continue
                assert not set(i_cls).issubset(
                    set(j_cls)
                ), f"Class {i_cls} is a subset of class {j_cls}."

        labset = set(self.labels)
        matches = [set(cls).issubset(labset) for cls in classes]

        if np.count_nonzero(matches) > 1:
            warnings.warn(f"More than one match for classes {classes}")
        else:
            try:
                id = np.where(matches)[0][0]
                self._group = id
                self._group_name = ", ".join(classes[id])
            except (TypeError, IndexError):
                pass

    def _merge(self, other):
        """TODO: Merge two calculators (to include additional SPIs)"""
        raise NotImplementedError()
        if self.identifier is not other.name:
            raise TypeError(f"Calculator name does do not match. Aborting merge.")

        for attr in ["name", "n_processes", "n_observations"]:
            selfattr = getattr(self.dataset, attr)
            otherattr = getattr(other.dataset, attr)
            if selfattr is not otherattr:
                raise TypeError(
                    f"Attribute {attr} does not match between calculators ({selfattr} != {otherattr})"
                )

    def get_stat_labels(self):
        """Get the labels for each statistic.

        Returns:
            stat_labels (dict): dictionary of
        """
        return {k: v.labels for k, v in zip(self._spis.keys(), self._spis.values())}

    def _get_correlation_df(self, with_labels=False, rmmin=False):
        # Sorts out pesky numerical issues in the unsigned spis
        if rmmin:
            self._rmmin()

        # Flatten (get Edge-by-SPI matrix)
        edges = self.table.stack()

        # Correlate the edge matrix (using pearson and/or spearman correlation)
        cf = pd.DataFrame(
            index=[c for c in edges.columns], columns=[c for c in edges.columns]
        )
        # Need to iterate through each pair to handle unsigned/signed statistics
        for i, s0 in enumerate(edges.columns):
            for j, s1 in enumerate(edges.columns[i + 1 :]):

                if self.spis[s0].issigned() and self.spis[s1].issigned():
                    # When they're both signed, just take the correlation
                    cf.iloc[[i, i + j + 1], [i, i + j + 1]] = edges[[s0, s1]].corr(
                        method="spearman"
                    )
                else:
                    # Otherwise, take the absolute value to make sure we compare like-for-like
                    cf.iloc[[i, i + j + 1], [i, i + j + 1]] = (
                        edges[[s0, s1]].abs().corr(method="spearman")
                    )

        cf.index.name = "SPI-1"
        cf.columns.name = "SPI-2"

        if with_labels:
            return cf, self.getstatlabels()
        else:
            return cf


def forall(func):
    def do(self, *args, **kwargs):
        try:
            for i in self._calculators.index:
                calc_ser = self._calculators.loc[i]
                for calc in calc_ser:
                    func(calc, *args, **kwargs)
        except AttributeError:
            raise AttributeError(
                f"No calculators in frame yet. Initialise before calling {func}"
            )

    return do


class CalculatorFrame:
    """ CalculatorFrame
        Container for batch level commands, like computing/pruning/initialising multiple datasets at once
    """
    def __init__(
        self,
        calculators=None,
        name=None,
        datasets=None,
        names=None,
        labels=None,
        **kwargs,
    ):
        if calculators is not None:
            self.set_calculator(calculators)

        self.name = name

        if datasets is not None:
            if names is None:
                names = [None] * len(datasets)
            if labels is None:
                labels = [None] * len(datasets)
            self.init_from_list(datasets, names, labels, **kwargs)

    @property
    def name(self):
        if hasattr(self, "_name") and self._name is not None:
            return self._name
        else:
            return ""

    @name.setter
    def name(self, n):
        self._name = n

    @staticmethod
    def from_calculator(calculator):
        cf = CalculatorFrame()
        cf.add_calculator(calculator)
        return cf

    def set_calculator(self, calculators):
        if hasattr(self, "_dataset"):
            Warning("Overwriting dataset without explicitly deleting.")
            del self._calculators

        if isinstance(calculators, Calculator):
            calculators = [calculators]

        if isinstance(calculators, CalculatorFrame):
            self.add_calculator(calculators)
        else:
            for calc in calculators:
                self.add_calculator(calc)

    def add_calculator(self, calc):

        if not hasattr(self, "_calculators"):
            self._calculators = pd.DataFrame()

        if isinstance(calc, CalculatorFrame):
            self._calculators = pd.concat([self._calculators.values, calc])
        elif isinstance(calc, Calculator):
            self._calculators = pd.concat(
                [self._calculators, pd.Series(data=calc, name=calc.name)],
                ignore_index=True,
            )
        elif isinstance(calc, pd.DataFrame):
            if isinstance(calc.iloc[0], Calculator):
                self._calculators = calc
            else:
                raise TypeError("Received dataframe but it is not in known format.")
        else:
            raise TypeError(f"Unknown data type: {type(calc)}.")

        self.n_calculators = len(self.calculators.index)

    def init_from_list(self, datasets, names, labels, **kwargs):
        base_calc = Calculator(**kwargs)
        for i, dataset in enumerate(datasets):
            calc = copy.deepcopy(base_calc)
            calc.load_dataset(dataset)
            calc.name = names[i]
            calc.labels = labels[i]
            self.add_calculator(calc)

    def init_from_yaml(
        self, document, detrend=True, normalise=True, n_processes=None, n_observations=None, **kwargs
    ):
        datasets = []
        names = []
        labels = []
        with open(document) as f:
            yf = yaml.load(f, Loader=yaml.FullLoader)

            for config in yf:
                try:
                    file = config["file"]
                    dim_order = config["dim_order"]
                    names.append(config["name"])
                    labels.append(config["labels"])
                    datasets.append(
                        Data(
                            data=file,
                            dim_order=dim_order,
                            name=names[-1],
                            detrend=detrend,
                            normalise=normalise,
                            n_processes=n_processes,
                            n_observations=n_observations,
                        )
                    )
                except Exception as err:
                    warnings.warn(f"Loading dataset: {config} failed ({err}).")

        self.init_from_list(datasets, names, labels, **kwargs)

    @property
    def calculators(self):
        """Return data array."""
        try:
            return self._calculators
        except AttributeError:
            return None

    @calculators.setter
    def calculators(self, cs):
        if hasattr(self, "calculators"):
            raise AttributeError(
                "You can not assign a value to this attribute"
                " directly, use the set_data method instead."
            )
        else:
            self._calculators = cs

    @calculators.deleter
    def calculators(self):
        warnings.warn("Overwriting existing calculators.")
        del self._calculators

    def merge(self, other):
        try:
            self._calculators = pd.concat(
                [self._calculators, other._calculators], ignore_index=True
            )
        except AttributeError:
            self._calculators = other._calculators

    @forall
    def compute(calc):
        calc.compute()

    @property
    def groups(self):
        groups = []
        for i in self._calculators.index:
            calc_ser = self._calculators.loc[i]
            for calc in calc_ser:
                groups.append(calc.group)
        return groups

    @forall
    def set_group(calc, *args):
        calc.set_group(*args)

    @forall
    def _rmmin(calc):
        calc._rmmin()

    def get_correlation_df(self, with_labels=False, **kwargs):
        if with_labels:
            mlabels = {}
            dlabels = {}

        shapes = pd.DataFrame()
        mdf = pd.DataFrame()
        for calc in [c[0] for c in self.calculators.values]:
            out = calc._get_correlation_df(with_labels=with_labels, **kwargs)

            s = pd.Series(
                dict(
                    n_processes=calc.dataset.n_processes,
                    n_observations=calc.dataset.n_observations,
                )
            )
            if calc.name is not None:
                s.name = calc.name
                shapes = pd.concat([shapes, pd.DataFrame(s).T])
            else:
                s.name = "N/A"
                shapes = pd.concat([shapes, pd.DataFrame(s).T])
            if with_labels:
                df = pd.concat({calc.name: out[0]}, names=["Dataset"])
                try:
                    mlabels = mlabels | out[1]
                except TypeError:
                    mlabels.update(out[1])
                dlabels[calc.name] = calc.labels
            else:
                df = pd.concat({calc.name: out}, names=["Dataset"])

            # Adds another hierarchical level giving the dataset name
            mdf = pd.concat([mdf, df])
        shapes.index.name = "Dataset"

        if with_labels:
            return mdf, shapes, mlabels, dlabels
        else:
            return mdf, shapes


class CorrelationFrame:
    def __init__(self, cf=None, **kwargs):
        self._slabels = {}
        self._dlabels = {}
        self._mdf = pd.DataFrame()
        self._shapes = pd.DataFrame()

        if cf is not None:
            if isinstance(cf, Calculator):
                cf = CalculatorFrame(cf)

            if isinstance(cf, CalculatorFrame):
                # Store the statistic-focused dataframe, statistic labels, and dataset labels
                (
                    self._mdf,
                    self._shapes,
                    self._slabels,
                    self._dlabels,
                ) = cf.get_correlation_df(with_labels=True, **kwargs)
                self._name = cf.name
            else:
                self.merge(cf)

    @property
    def name(self):
        if not hasattr(self, "_name"):
            return ""
        else:
            return self._name

    @name.setter
    def name(self, n):
        self._name = n

    @property
    def shapes(self):
        return self._shapes

    @property
    def mdf(self):
        return self._mdf

    @property
    def ddf(self):
        if not hasattr(self, "_ddf") or self._ddf.size != self._mdf.size:
            self._ddf = convert_mdf_to_ddf(self.mdf)
        return self._ddf

    @property
    def n_datasets(self):
        return self.ddf.shape[1]

    @property
    def n_spis(self):
        return self.mdf.shape[1]

    @property
    def mlabels(self):
        return self._slabels

    @property
    def dlabels(self):
        return self._dlabels

    @mdf.setter
    def mdf(self):
        raise AttributeError("Do not directly set the mdf attribute.")

    @mlabels.setter
    def mlabels(self):
        raise AttributeError("Do not directly set the mlabels attribute.")

    @dlabels.setter
    def dlabels(self):
        raise AttributeError("Do not directly set the dlabels attribute.")

    def merge(self, other):
        if not all(isinstance(i[0], str) for i in self._mdf.index):
            raise TypeError(
                f"This operation only works with named calculators (set each calc.name property)."
            )

        try:
            self._ddf = self.ddf.join(other.ddf)
            self._mdf = pd.concat([self._mdf, other.mdf], verify_integrity=True)
            self._shapes = pd.concat(
                [self._shapes, other.shapes], verify_integrity=True
            )
        except KeyError:
            self._ddf = copy.deepcopy(other.ddf)
            self._mdf = copy.deepcopy(other.mdf)
            self._shapes = copy.deepcopy(other.shapes)

        try:
            self._slabels = self._slabels | other.mlabels
            self._dlabels = self._dlabels | other.dlabels
        except TypeError:
            self._slabels.update(other.mlabels)
            self._dlabels.update(other.dlabels)

    def get_pvalues(self):
        if not hasattr(self, "_pvalues"):
            n = self.shapes["n_observations"]
            nstats = self.mdf.shape[1]
            ns = np.repeat(n.values, nstats**2).reshape(
                self.mdf.shape[0], self.mdf.shape[1]
            )
            rsq = self.mdf.values**2
            fval = ns * rsq / (1 - rsq)
            self._pvalues = stats.f.sf(fval, 1, ns - 1)
        return pd.DataFrame(
            data=self._pvalues, index=self.mdf.index, columns=self.mdf.columns
        )

    def compute_significant_values(self):
        pvals = self.get_pvalues()
        nstats = self.mdf.shape[1]
        self._insig_ind = pvals > 0.05 / nstats / (nstats - 1) / 2

        if not hasattr(self, "_insig_group"):
            pvals = pvals.droplevel(["Dataset", "Type"])
            group_pvalue = pd.DataFrame(
                data=np.full([pvals.columns.size] * 2, np.nan),
                columns=pvals.columns,
                index=pvals.columns,
            )
            for f1 in pvals.columns:
                print(f"Computing significance for {f1}...")
                for f2 in [
                    f
                    for f in pvals.columns
                    if f is not f1 and np.isnan(group_pvalue[f1][f])
                ]:
                    cp = pvals[f1][f2]
                    group_pvalue[f1][f2] = stats.combine_pvalues(cp[~cp.isna()])[1]
                    group_pvalue[f2][f1] = group_pvalue[f1][f2]
            self._insig_group = group_pvalue > 0.05

    def get_average_correlation(
        self, thresh=0.2, absolute=True, summary="mean", remove_insig=False
    ):
        mdf = copy.deepcopy(self.mdf)

        if absolute:
            ss_adj = getattr(mdf.abs().groupby("SPI-1"), summary)()
        else:
            ss_adj = getattr(mdf.groupby("SPI-1"), summary)()
        ss_adj = (
            ss_adj.dropna(thresh=ss_adj.shape[0] * thresh, axis=0)
            .dropna(thresh=ss_adj.shape[1] * thresh, axis=1)
            .sort_index(axis=1)
        )
        if remove_insig:
            ss_adj[self._insig_group.sort_index()] = np.nan

        return ss_adj

    def get_feature_matrix(self, sthresh=0.8, dthresh=0.2, dropduplicates=True):

        fm = self.ddf
        if dropduplicates:
            fm = fm.drop_duplicates()

        # Drop datasets based on NaN threshold
        num_dnans = dthresh * fm.shape[0]
        fm = fm.dropna(axis=1, thresh=num_dnans)

        # Drop measures based on NaN threshold
        num_snans = sthresh * fm.shape[1]
        fm = fm.dropna(axis=0, thresh=num_snans)
        return fm

    @staticmethod
    def _verify_classes(classes):
        # Ensure this is a list of lists
        for i, cls in enumerate(classes):
            if not isinstance(cls, list):
                classes[i] = [cls]

        for i, i_cls in enumerate(classes):
            for j, j_cls in enumerate(classes):
                if i == j:
                    continue
                assert not set(i_cls).issubset(
                    set(j_cls)
                ), f"Class {i_cls} is a subset of class {j_cls}."

    @staticmethod
    def _get_group(labels, classes, instance, verbose=False):
        labset = set(labels)
        matches = [set(cls).issubset(labset) for cls in classes]

        # Iterate through all
        if np.count_nonzero(matches) > 1:
            if verbose:
                print(
                    f"More than one match in for {instance} whilst searching for {classes} within {labels}). Choosing first one."
                )

        try:
            myid = np.where(matches)[0][0]
            return myid
        except (TypeError, IndexError):
            if verbose:
                print(f"{instance} has no match in {classes}. Options are {labels}")
            return -1

    @staticmethod
    def _set_groups(classes, labels, group_names, group):
        CorrelationFrame._verify_classes(classes)
        for m in labels:
            group[m] = CorrelationFrame._get_group(labels[m], classes, m)

    def set_sgroups(self, classes):
        # Initialise the classes
        self._sgroup_names = {i: ", ".join(c) for i, c in enumerate(classes)}
        self._sgroup_names[-1] = "N/A"

        self._sgroup_ids = {m: -1 for m in self._slabels}
        CorrelationFrame._set_groups(
            classes, self._slabels, self._sgroup_names, self._sgroup_ids
        )

    def set_dgroups(self, classes):
        self._dgroup_names = {i: ", ".join(c) for i, c in enumerate(classes)}
        self._dgroup_names[-1] = "N/A"

        self._dgroup_ids = {d: -1 for d in self._dlabels}
        CorrelationFrame._set_groups(
            classes, self._dlabels, self._dgroup_names, self._dgroup_ids
        )

    def get_dgroup_ids(self, names=None):
        if names is None:
            names = self._ddf.columns

        return [self._dgroup_ids[n] for n in names]

    def get_dgroup_names(self, names=None):
        if names is None:
            names = self._ddf.columns

        return [self._dgroup_names[i] for i in self.get_dgroup_ids(names)]

    def get_sgroup_ids(self, names=None):
        if names is None:
            names = self._mdf.columns

        return [self._sgroup_ids[n] for n in names]

    def get_sgroup_names(self, names=None):
        if names is None:
            names = self._mdf.columns

        return [self._sgroup_names[i] for i in self.get_sgroup_ids(names)]

    def relabel_spis(self, names, labels):
        assert len(names) == len(labels), "Length of spis must equal length of labels."

        for n, l in zip(names, labels):
            try:
                self._slabels[n] = l
            except AttributeError:
                self._slabels = {n: l}

    def relabel_data(self, names, labels):
        assert len(names) == len(
            labels
        ), "Length of datasets must equal length of labels."

        for n, l in zip(names, labels):
            self._dlabels[n] = l
